#include "g_local.h"
#include "BotMain.h"


/***************************************************************************************
original code: TeamWhore

filename: BotMove.c

Description: This file contains the functions providing help with movement

###################### Function Descriptions ##################################

qboolean CanMove(gentity_t *bot, int dir)
   Checks if the bot can move in the given direction

void BotMove(gentity_t *bot, int dir, usercmd_t *ucmd, int flags)
  This simply takes the direction the bot is moving, how the bot is moving (forward, backwards..)
  and adjusts the 'user cmd' accordingly

qboolean BotMoveNearerNode(gentity_t *bot)
  Tries to move a bot closer to the node he is at, to compensate for nodes placed very near obstacles

qboolean BotSpecialMove(gentity_t *bot)
  Attempts to have a bot move in whatever way he can, to get out from being stuck.  So far basically just
	tries crouching or jumping

void BotFollowPlayer(gentity_t *bot)
  This function handles when the bot is trying to follow a player who has issued a 'follow me' voice chat

int BotFindFreeDirection(gentity_t *bot)
  Minimal search to find an open direction around the bot, so he can give himself health/ammo

###############################################################################

-----changes:---------------------
  March 4, 2004 
	rince - added comments to help explain the code
  March 4-5, 2004
    TeamWhore - merged rince's comments with mine
	    added 'sprint' support to BotMove
		added function BotFollowPlayer to handle bot's response to a 'follow me' voice chat
  March 5, 2004
	TeamWhore - moved MyInfront function to BotMisc.c
		moved BotFindFreeDirection here from BotMisc.c
  April 2, 2004
	TeamWhore - changed CanMove to use better content-masking and distances
		also changed CanMove from point to box collision detection
  April-July, 2004
    TeamWhore - updated CanMove to use better-sized bounding boxes for collision detection
		also changed CanMove to check directly in front of the bot as well, in case he is facing up or down
		added function BotMoveNearerNode to allow bots to approach their current node to get by possible obstacles
		added function BotSpecialMove to try to resolve situations where a bot simply can't move forward

****************************************************************************************/


/*////////////////////////////////////////////
CanMove
performs a very rudimentary check to see if the bot can move
in a specified direction
*/////////////////////////////////////////////
qboolean CanMove(gentity_t *bot, int dir)
{
	vec3_t angles, forward, start, end, mins, maxs;
	//used in the actual collision check
	trace_t tr;

	//copy the angles the bot is facing
	VectorCopy(bot->s.angles, angles);
	
	//change angles based on what direction the bot wants to move (in degrees)
	if(dir == DIR_LEFT)
		angles[1] += 90;
	else if(dir == DIR_RIGHT)
		angles[1] -= 90;
	else if(dir == DIR_BACKWARD)
		angles[1] -=180;
	
	//set 'forward' to the direction the bot wants to move
	AngleVectors (angles, forward, NULL, NULL);
	//set 'start' to the bot's current position
	VectorCopy(bot->r.currentOrigin, start);
	//take into account the bot's view height
	//start[2] += bot->client->ps.viewheight;	//BOTS TW - commented this out as a test...seems to work better without it

	//set 'end' to 32 units 'forward' of 'start' - from 'start', 32 units in the dir the bot wants to move
	//we don't want to set this much higher, or the bot may often think he is blocked at a node when he actually isn't
	VectorMA(start, 32, forward, end);

	VectorCopy(bot->client->ps.mins, mins);
	VectorCopy(bot->client->ps.maxs, maxs);

	if (!(botinfo[bot->s.number].moveFlags & PATH_CROUCH))
	{
		//this basically alters the collision bounding box of the bot to not include things low to the ground
		//I do this so that bots are less likely to think they are 'blocked' when they're simply on a hill
		mins[2] = 0;
	}
	else
	{
		//if the bot is crouching, leave the mins, but set the maxs to recognize the bot's compact form
		maxs[2] = bot->client->ps.crouchMaxZ;
	}
	
	//perform the actual collision check to see if the bot would hit a player or a solid object
	trap_Trace(&tr, start, mins, maxs, end, bot->s.number, MASK_PLAYERSOLID);

	if (tr.fraction != 1.0)			//hit something - tr.fraction is 1.0 if nothing was hit
	{
		//if we hit something, check if the bot can move straight forward (even if he's facing up or down)
		//first recalculate all the vectors
		angles[0] = 0;
		AngleVectors(angles, forward, NULL, NULL);
		VectorMA(start, 32, forward, end);

		//do the new trace
		trap_Trace(&tr, start, mins, maxs, end, bot->s.number, MASK_PLAYERSOLID);

		//if this one fails too, we're saying we can't move :)
		if (tr.fraction != 1.0)
			return qfalse;
	}

	return qtrue;					//can move
}


/*////////////////////////////////////////
Moves the bot in the specified direction, accommodating for flags
does not perform checking to see if movement is possible
*/////////////////////////////////////////
void BotMove(gentity_t *bot, int dir, int flags)
{
	switch (dir)
	{
	case DIR_FORWARD:				//moving forward
		botinfo[bot->s.number].ucmd->forwardmove += 127;
		break;
	case DIR_RIGHT:					//strafing right
		botinfo[bot->s.number].ucmd->rightmove += 127;
		break;
	case DIR_BACKWARD:				//moving backward
		botinfo[bot->s.number].ucmd->forwardmove += -127;	//note the negative sign
		break;	
	case DIR_LEFT:					//strafing left
		botinfo[bot->s.number].ucmd->rightmove += -127;		//note the negative sign
		break;
	default:
		break;
	}

	if (flags & BOTMOVE_CROUCH)								//if the bot is crouching
	{
		if (bot->waterlevel < 2)							//don't crouch in the water
			botinfo[bot->s.number].ucmd->upmove -= 127;		//begin crouch
	}

	if (flags & BOTMOVE_JUMP)			//if the bot is jumping
	{
		if (bot->client->ps.groundEntityNum != ENTITYNUM_NONE) //if i'm not already in the air
			botinfo[bot->s.number].ucmd->upmove += 127;			//jump!
	}

	if (flags & BOTMOVE_WALK)									//if the bot is walking
		botinfo[bot->s.number].ucmd->buttons |= BUTTON_WALKING;	//press the 'walk' button - caps lock for people

	if (flags & BOTMOVE_SPRINT)									//if the bot is sprinting
		botinfo[bot->s.number].ucmd->buttons |= BUTTON_SPRINT;	//press the 'sprint' button - shift for people
}

/*//////////////////////////////////////////////
BotMoveNearerNode
This function just attempts to have a bot move closer
to his current node that he normally would.  This basically
tries to resolve MINOR blocking issues caused by poor node
placement or obstacles very near a node
*///////////////////////////////////////////////
qboolean BotMoveNearerNode(gentity_t *bot)
{
	vec3_t v;
	float dist;

	if (botinfo[bot->s.number].currentNode == NODE_INVALID)
		return qfalse;

	//check this flag so the bot won't get 'stuck' on the node
	if (botinfo[bot->s.number].onNode == qtrue)
		return qfalse;

	VectorSubtract(nodes[botinfo[bot->s.number].currentNode].pos, bot->r.currentOrigin, v);
	v[2] = 0;	//drop out the vertical component here, since the node's height is pretty irrelevant
	dist = VectorLength(v);

	if (dist > 32)
		return qfalse;

	if (dist < 5)
	{
		botinfo[bot->s.number].onNode = qtrue;
		return qfalse;
	}

	vectoangles(v, bot->s.angles);
	BotMove(bot, DIR_FORWARD, BOTMOVE_CROUCH);

	return qtrue;
}

/*//////////////////////////////////////////////
BotSpecialMove
Called when a bot's normal movement is blocked in some way.
Checks if various non-normal movements are open, and executes
them if they are
*///////////////////////////////////////////////
qboolean BotSpecialMove(gentity_t *bot)
{
	vec3_t angles, forward, start, end, maxs;
	trace_t tr;

	VectorCopy(bot->s.angles, angles);
	AngleVectors(angles, forward, NULL, NULL);
	
	VectorCopy(bot->r.currentOrigin, start);

	//first check for crouching
	//start[2] += bot->client->ps.crouchViewHeight;
	VectorMA(start, 36, forward, end);
	VectorCopy(bot->client->ps.maxs, maxs);
	maxs[2] = bot->client->ps.crouchMaxZ;

	//don't need to check this if we're already ducked!
	if (!(bot->client->ps.pm_flags & PMF_DUCKED))
	{
		trap_Trace(&tr, start, bot->client->ps.mins, maxs, end, bot->s.number, MASK_PLAYERSOLID);
		if (tr.fraction == 1.0)
		{
			BotMove(bot, DIR_FORWARD, BOTMOVE_CROUCH);
			return qtrue;
		}
	}

	//next check for jumping
	if (bot->client->ps.groundEntityNum != ENTITYNUM_NONE)		//can't jump if already in the air
	{
		//first see if we CAN jump up (no low ceiling etc)
		maxs[2] = bot->client->ps.maxs[2] + 12;	//this 12 is a random guess...how high do you jump?
		trap_Trace(&tr, bot->client->ps.origin, bot->client->ps.mins, maxs, bot->client->ps.origin, bot->s.number, MASK_PLAYERSOLID);

		if (!tr.allsolid)	//I hope this means we can jump...basically taken from PM_CheckDuck
		{
			//check if jumping would allow us to move forward
			start[2] = bot->r.currentOrigin[2] + bot->client->ps.standViewHeight + 12;
			VectorMA(start, 36, forward, end);
			trap_Trace(&tr, start, bot->client->ps.mins, bot->client->ps.maxs, end, bot->s.number, MASK_PLAYERSOLID);
			if (tr.fraction == 1.0)
			{
				BotMove(bot, DIR_FORWARD, BOTMOVE_JUMP);
				return qtrue;
			}
		}
	}

	//didn't resolve a move here
	return qfalse;
}


/*//////////////////////////////////////////////
BotFollowPlayer
Function to follow player who issued a 'follow me' command
*///////////////////////////////////////////////
void BotFollowPlayer(gentity_t *bot)
{
	vec3_t v;
	float dist;
	//get the player we are following
	gentity_t* player = &g_entities[botinfo[bot->s.number].followingNum];

	//don't follow someone who isn't on the server or who's dead...
	if (!player || player->client->ps.pm_type == PM_DEAD || player->client->pers.connected == CON_DISCONNECTED)
	{
		//reset these because the bot isn't following anybody anymore
		botinfo[bot->s.number].followingNum = ENTITYNUM_NONE;
		botinfo[bot->s.number].followingTime = 0;
		//end the function
		return;
	}

	//get the direction and distance from the bot to the player
	VectorSubtract(player->r.currentOrigin, bot->r.currentOrigin, v);
	dist = VectorNormalize(v);
	//face the player
	vectoangles(v, bot->s.angles);

	//if the bot is more than 140 units away from the player, move towards him (otherwise don't move)
	if (dist > 140)
		BotMove(bot, DIR_FORWARD, 0);
}


/*//////////////////////////////////////////////////////////////////
BotFindFreeDirection
This function checks in a small radius around the bot to see which
direction is open.  It checks in the order forward, right, left, backward
Fairly rudimentary
*///////////////////////////////////////////////////////////////////
int BotFindFreeDirection(gentity_t *bot)
{
	trace_t tr;
	vec3_t start, end, forward, right;
	int dir = DIR_NOMOVE;

	//this function gets the vectors at right angles relative to the direction you're facing (directly in front of you, 90 degrees to the right, and straight up)
	AngleVectors(bot->s.angles, forward, right, NULL);
	VectorCopy(bot->r.currentOrigin, start);			//start from bot's position
	VectorMA(start, 300, forward, end);					//go forward 300 units

	//check to see if there's anything/anyone in the way
	trap_Trace(&tr, start, NULL, NULL, end, bot->s.number, MASK_PLAYERSOLID);

	if (tr.fraction < 1.0)		//if there's a person/object within 300 units forward
	{
		VectorMA(start, 300, right, end);		//check to the right
		trap_Trace(&tr, start, NULL, NULL, end, bot->s.number, MASK_PLAYERSOLID);

		if (tr.fraction < 1.0)		//if there's someone to the right as well
		{
			vec3_t left;
			VectorNegate(right, left);		//opposite of right is left :P
			VectorMA(start, 300, left, end);	//check to the left
			trap_Trace(&tr, start, NULL, NULL, end, bot->s.number, MASK_PLAYERSOLID);

			if (tr.fraction < 1.0)	//if to the left is blocked too
			{
				vec3_t backward;
				VectorNegate(forward, backward);	//check backwards
				VectorMA(start, 300, backward, end);
				trap_Trace(&tr, start, NULL, NULL, end, bot->s.number, MASK_PLAYERSOLID);

				if (tr.fraction < 1.0)		//if this is blocked too...at this point, all 4 directions are blocked
					return DIR_NOMOVE;		//this shouldn't happen much unless they all bunch up
				else
					dir = DIR_BACKWARD;		//behind the bot is open
			}
			else
				dir = DIR_LEFT;				//to the left of the bot is open
		}
		else
			dir = DIR_RIGHT;				//to the right of the bot is open
	}
	else
		dir = DIR_FORWARD;					//in front of the bot is open

	return dir;								//return the open direction
}
