#include "g_local.h"
#include "BotMain.h"


/***************************************************************************************
original code: TeamWhore

filename: BotNodes.c

Description: This file contains the functions providing Node functions

###################### Function Descriptions ##################################

qboolean AddNode(vec3_t origin, int fl, short int* ents, int objFl)
	adds a node using given point, flags, entities, objective flags
	the entities are any map objects associated with the node
	return qtrue if the procedure was successful

qboolean ConnectNodes(int from, int to, int flags)
	connects node 'from' to node 'to' using path flags from 'flags'
	returns qtrue if the procedure was successful

void ClearNodes()
	Clears all nodes from our 'nodes' array

void InitNodes()
	Initialize nodes by attempting to load the nodes file that is the same as the map
	(if loading mp_beach, tries to load mp_beach.nod)

short int BotFindClosestNode(gentity_t *bot)
	Finds the closest node near the given entity 'bot'
	Returns the 'number' of the node that belongs in the 'nodes' array
	return value of 15 means 'nodes [15]'

short int BotGetNextNode(gentity_t *bot, int *state)
	Get the next node the bot should move to
	return value of 440, means nodes[440]

qboolean LoadNodes(char *filename)
	Loads nodes for the map using the given filename

void SaveNodes(char *filename)
	saves nodes using given filename

void BotFindLongTermGoal(gentity_t *bot)
	searches for a longterm objective for the bot to try for
	does not interfere with short-range goals

void UpdateNodes()
	called every frame to update the unreachable/objective flags of
		any nodes for which these have been cleared

###############################################################################

-----changes:---------------------
  March 4, 2004 
	rince - added comments to help explain the code
  March 4-5, 2004
    TeamWhore - merged rince's comments with my own
	    removed lines in BotGetNextNode that did useless node selection
  March 5, 2004
    TeamWhore - changed BotGetNextNode just to allow for movement, still not purposeful though
  March 6, 2004
	TeamWhore - added new flags initialization to AddNode and ClearNodes (for parsed objective flags)
		heavily altered BotGetNextNode to allow for more purposeful movement, but still not finished
		added BotFindLongTermGoal function to allow the bot to have a future objective to work towards
		added GetCost function to determine total distance the bot would have to travel to reach a certain node
  March 7-8, 2004
	TeamWhore - updated BotGetNextNode, BotFindLongTermGoal and GetCost to work a little better
		fixed crashes relating to invalid nodes and infinite looping
		changed LoadNodes and SaveNodes to accept new .nod file format ('Ubersoldat' instead of 'tardbot')
  March 9, 2004
	TeamWhore - fixed bug in GetCost which would prevent a bot from moving onto their longterm goal node
		added code to BotFindLongTermGoal so a bot's goal choice will be more intelligent
		added function UpdateNodes to ensure that all the node flags are properly set/cleared each frame
  April 2, 2004
	TeamWhore - changed BotGetClosestNode to correctly handle 'blocked' nodes
		moved GetCost (and other funcs?) to new file BotUtils.c
		completely reworked BotGetNextNode to handle new pathfinding/goal movement
		added code to BotGetNextNode to handle BOTSTATE_WANDER movement
		added 'state' param to BotGetNextNode function definition to allow for wander movement
  April-July, 2004
	TeamWhore - changed BotGetNextNode once again to remove references to GetCost
		updated node loading/saving code to reflect new version 2 of .nod files
		small bug fixes

****************************************************************************************/


/*//////////////////////////////////////////////
AddNode
creates a new waypoint/node with the specified values
*//////////////////////////////////////////////
qboolean AddNode(vec3_t origin, int fl, short int* ents, int objFl)
{
	int objTeam;
	//make sure we don't already have too many nodes
	//curently MAX_NODES is 1024
	if (numNodes+1 > MAX_NODES)
		return qfalse;

	VectorCopy(origin, nodes[numNodes].pos);	//set the node's position
	nodes[numNodes].num = numNodes;				//set it's number
	nodes[numNodes].flags = fl;					//set the flags (NODE_OBJECTIVE, for example)
	nodes[numNodes].objectNum[0] = ents[0];		//set any map objects associated with this node
	nodes[numNodes].objectNum[1] = ents[1];		//only applies to objects linked to the unreachable flag
	nodes[numNodes].objectNum[2] = ents[2];
	nodes[numNodes].objFlags = objFl;			//set the objective flags

	//get some objective-related data out of the flags (this is based on the .wps format - see the forum)
	nodes[numNodes].objEntity = ((objFl & 0x0FFF0) >> 7);
	nodes[numNodes].objType = ((objFl & 0x0000F) >> 2);
	
	objTeam = (objFl & 1);
	if (objTeam == 0)
		nodes[numNodes].objTeam = TEAM_BLUE;
	else if (objTeam == 1)
		nodes[numNodes].objTeam = TEAM_RED;

	numNodes++;

	return qtrue;
}


/*////////////////////////////////////////////////
ConnectNodes
Connects 2 nodes and sets the flags for the path between them
/*////////////////////////////////////////////////
qboolean ConnectNodes(int from, int to, int flags)
{
	//check that we don't have too many connections from the 'from' node already
	if (nodes[from].numLinks+1 > MAX_LINKS)
		return qfalse;

	//check that we are creating a path between 2 valid nodes
	if ((nodes[from].flags == NODE_INVALID) || (to == NODE_INVALID))	//nodes[to].type is invalid on LoadNodes()
		return qfalse;

	//update the individual nodes
	nodes[from].links[nodes[from].numLinks].target = to;
	nodes[from].links[nodes[from].numLinks].flags = flags;

	nodes[from].numLinks++;

	return qtrue;
}


/*////////////////////////////////////////////////
ClearNodes
Clears all the nodes and links to unused/invalid values
*/////////////////////////////////////////////////
void ClearNodes()
{
	int i, j;

	for (i = 0; i < MAX_NODES; i++)
	{
		nodes[i].num = -1;
		nodes[i].pos[0] = nodes[i].pos[1] = nodes[i].pos[2] = 0.0;
		nodes[i].flags = NODE_INVALID;
		nodes[i].objectNum[0] = nodes[i].objectNum[1] = nodes[i].objectNum[2] = ENTITYNUM_NONE;
		nodes[i].objFlags = 0;
		nodes[i].objTeam = 0;
		nodes[i].objType = 0;
		nodes[i].objEntity = ENTITYNUM_NONE;
		nodes[i].numLinks = 0;

		for (j = 0; j < nodes[i].numLinks; j++)
		{
			nodes[i].links[j].flags = 0;
			nodes[i].links[j].target = NODE_INVALID;
		}
	}

	numNodes = 0;
}


/*///////////////////////////////////////////////
InitNodes
Called in G_InitGame, clears all the nodes/links
and tries to load in the default node file for the map
which is <mapname>.nod - so for mp_beach, it loads mp_beach.nod
*////////////////////////////////////////////////
void InitNodes()
{
	fileHandle_t f;
	vmCvar_t mapname;
	char filename[64];

	//clear out all the node vales
	ClearNodes();

	//these lines set 'filename' to <mapname>.nod
	trap_Cvar_Register(&mapname, "mapname", "", CVAR_ROM|CVAR_SERVERINFO);
	strcpy(filename, mapname.string);
	strcat(filename, ".nod");

	trap_FS_FOpenFile(filename, &f, FS_READ);
	//if we can't open the file, just display an error message
	if (!f)
	{
		G_Printf("Couldn't load waypoints for map %s\n", mapname.string);
		return;
	}

	trap_FS_FCloseFile(f);

	//load the nodes for this map from the file
	LoadNodes(filename);
}


/*////////////////////////////////////////////////
BotFindClosestNode
loops through all the nodes and returns the one that is
closest to the bot's position and that doesn't have the 'unreachable' flag set
*/////////////////////////////////////////////////
short int BotFindClosestNode(gentity_t *bot)
{
	int			i;
	float		closest = -1;
	float		dist;
	short int	node = NODE_INVALID;
	vec3_t		v;
	trace_t		tr;


	for(i = 0; i < numNodes; i++)
	{
		if (nodes[i].flags == NODE_INVALID)		//change this to AND? - since flags values are now logical-ORed together...
			continue;
		if ((bot->client->sess.sessionTeam == TEAM_RED) && (nodes[i].flags & NODE_AXIS_UNREACHABLE))
			continue;
		if ((bot->client->sess.sessionTeam == TEAM_BLUE) && (nodes[i].flags & NODE_ALLY_UNREACHABLE))
			continue;

		//find the distance to the node we are checking
		VectorSubtract(nodes[i].pos, bot->r.currentOrigin,v);
		dist = VectorLength(v);

		//if I'm already at this node, don't return it as the closest one
		if (dist <= 36)
			continue;

		//skip if we can't get to it (eg, was unreachable, but one of ents is blown, but this one still isn't)
		//eg - beach door blown, we're at wall, don't try to go through it
		trap_Trace(&tr, bot->r.currentOrigin, NULL, NULL, nodes[i].pos, bot->s.number, MASK_DEADSOLID);
		if (tr.fraction != 1.0)
			continue;
		
		//if this is the current closest node
		if((dist < closest) || (closest == -1))
		{
			//set it as the closest node
			node = i;
			closest = dist;
		}
	}
	
	//return the closest reachable node
	return node;
}


/*///////////////////////////////////////////////////
BotGetNextNode
if the bot has reached a node, this function selects the next node
that he will go to, and returns it
right now it's being developed, feel free to experiment
*////////////////////////////////////////////////////
short int BotGetNextNode(gentity_t *bot, int *state)
{
	if (*state == BOTSTATE_WANDER)
	{
		//wander movement - doesn't have a goal, just moves to the first connected node
		//and tries to avoid moving back and forth if possible
		//note that this is a VERY poor way to handle this - bots will likely often run in circles if possible
		int i;
		node_t *cur = &nodes[botinfo[bot->s.number].currentNode];	//the node the bot has just reached
		node_t *nd = NULL;

		if ((cur->flags == NODE_INVALID) || (cur->num == NODE_INVALID))		//if he's not at a node, just find the closest one
			return BotFindClosestNode(bot);

		for (i = 0; i < cur->numLinks; i++)			//loop through all the nodes connected to this node
		{
			nd = &(nodes[cur->links[i].target]);

			//skip this link if the connected node is invalid or currently unreachable for the bot's team
			if ((nd->flags == NODE_INVALID) || (nd->num == NODE_INVALID))
				continue;
			if ((nd->flags & NODE_ALLY_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_BLUE))
				continue;
			if ((nd->flags & NODE_AXIS_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_RED))
				continue;
			if (cur->links[i].flags & PATH_BLOCKED)		//if this path is blocked, skip it
				continue;

			//if this is the last run through the loop and we still haven't found a good node, return this one
			//this makes sure that something is returned
			if (i == cur->numLinks-1)
				return nd->num;

			//don't move back to the previous node (if we have to, the above line handles it)
			if (nd->num == botinfo[bot->s.number].previousNode)
				continue;

			return nd->num;		//just return the first non-backwards link
		}

		return NODE_INVALID;	//didn't find a link (does this happen?), so return invalid
	}
	else	//BOTSTATE_MOVE
	{	
		short int node = NODE_INVALID;

		//we should never call this in BOTSTATE_MOVE with no goal
		//setup the goal/path in HandleIdleState
		if (botinfo[bot->s.number].longTermGoal == NODE_INVALID)
		{
			*state = BOTSTATE_STAND;
			return NODE_INVALID;
		}

		if (botinfo[bot->s.number].pathsize == 0)	//if the bot is at the end of his path, this shouldn't have been called
		{
			*state = BOTSTATE_STAND;	//reset the state
			botinfo[bot->s.number].longTermGoal = NODE_INVALID;	//reset to having no goal
			return NODE_INVALID;
		}

		node = botinfo[bot->s.number].pathlist[botinfo[bot->s.number].pathsize-1];	//pathlist is in reverse order
		botinfo[bot->s.number].pathsize--;	//mark that we've moved another node

		return node;
	}
}


/*/////////////////////////////////////////////
LoadNodes
Initializes the nodes for a map from a file
The file must be in a .pk3 file in the mod folder
This function is completely dependent on the file format for the node files
as well as the format of the node_t structure
*//////////////////////////////////////////////
qboolean LoadNodes(char *filename)
{
	fileHandle_t f;
	int i, j;

	short int num, flags, objNum[3] = {0, 0, 0}, objFlags, numLinks;
	vec3_t vec;
	short int target, fl2;

	char name[] = BOT_MOD_NAME;
	char nm[64] = "";
	float version;
	vmCvar_t mapname;
	char map[64] = "";
	char mp[64] = "";

	short int numberNodes;
	

	//open the node file for reading, return false on error
	trap_FS_FOpenFile(filename, &f, FS_READ);
	if (!f)
	{
		G_Printf("Reading from %s failed\n", filename);
		return qfalse;
	}

	trap_Cvar_Register(&mapname, "mapname", "", CVAR_ROM|CVAR_SERVERINFO);	//get the map name
	strcpy(mp, mapname.string);


	trap_FS_Read(&nm, strlen(name)+1, f);	//read in a string the size of the mod name (+1 is because all strings end in hex '00')
	if (Q_stricmp(nm, name) != 0)
	{
		G_Printf("Incompatible or corrupt node file, load aborted\n");
		trap_FS_FCloseFile(f);
		return qfalse;			//if the file doesn't start with the mod name, it's not a valid .nod file, so abort
	}

	trap_FS_Read(&version, sizeof(float), f);		//read and make sure the version is the same
	if (version != UBERSOLDAT_VERSION)
	{
		G_Printf("Node file is wrong version, load aborted\n");
		trap_FS_FCloseFile(f);
		return qfalse;
	}

	trap_FS_Read(&map, strlen(mp)+1, f);		//make sure the file is for the current map
	if (Q_stricmp(map, mp) != 0)
	{
		G_Printf("Node file is for a different map, load aborted\n");
		trap_FS_FCloseFile(f);
		return qfalse;
	}

	trap_FS_Read(&numberNodes, sizeof(short int), f);	//read in the number of nodes in the map

	for (i = 0; i < numberNodes; i++)				//loop through all the nodes
	{
		//read in all the node info stored in the file
		trap_FS_Read(&num, sizeof(short int), f);			//used for links, but not written directly to nodes
		trap_FS_Read(&vec, sizeof(vec3_t), f);
		trap_FS_Read(&flags, sizeof(short int), f);
		trap_FS_Read(objNum, sizeof(short int)*3, f);
		trap_FS_Read(&objFlags, sizeof(int), f);
		trap_FS_Read(&numLinks, sizeof(short int), f);

		AddNode(vec, flags, objNum, objFlags);			//add the node

		//loop through all of the links and read the data
		for (j = 0; j < numLinks; j++)
		{
			trap_FS_Read(&target, sizeof(short int), f);
			trap_FS_Read(&fl2, sizeof(short int), f);

			ConnectNodes(num, target, fl2);				//add any links
		}
	}

	trap_FS_FCloseFile(f);		//close the file

	G_Printf("Successfully read %s\n", filename);	//report that it worked

	return qtrue;
}


/*///////////////////////////////////////////////////
SaveNodes
saves all the current nodes and links to a file
default filename is <mapname>.nod
This function depends on the file format for .nod files
and the node_t structure format
*////////////////////////////////////////////////////
void SaveNodes(char *filename)
{
	fileHandle_t f;
	int i;
	short int j;
	float version = UBERSOLDAT_VERSION;		//version is 1.0 for now
	char name[] = BOT_MOD_NAME;
	vmCvar_t mapname;
	char map[64] = "";

	//try to open the output file, return if it failed
	trap_FS_FOpenFile(filename, &f, FS_WRITE);
	if (!f)
	{
		G_Printf("Error opening node file\n");
		return;
	}

	trap_Cvar_Register(&mapname, "mapname", "", CVAR_SERVERINFO|CVAR_ROM);	//get the map name
	strcpy(map, mapname.string);

	trap_FS_Write(&name, strlen(name)+1, f);	//write the mod name to the file
	trap_FS_Write(&version, sizeof(float), f);	//write the version of this file
	trap_FS_Write(&map, strlen(map)+1, f);		//write the map name

	trap_FS_Write(&numNodes, sizeof(short int), f);	//write the number of nodes in the map

	for (i = 0; i < numNodes; i++)		//loop through all the nodes
	{
		//write all the node data to the file
		trap_FS_Write(&(nodes[i].num), sizeof (short int), f);
		trap_FS_Write(&(nodes[i].pos), sizeof (vec3_t), f);
		trap_FS_Write(&(nodes[i].flags), sizeof (short int), f);
		trap_FS_Write(&(nodes[i].objectNum), sizeof(short int)*3, f);
		trap_FS_Write(&(nodes[i].objFlags), sizeof(int), f);
		trap_FS_Write(&(nodes[i].numLinks), sizeof(short int), f);

		for (j = 0; j < nodes[i].numLinks; j++)
		{
			trap_FS_Write(&(nodes[i].links[j].target), sizeof(short int), f);
			trap_FS_Write(&(nodes[i].links[j].flags), sizeof(short int), f);
		}
	}

	trap_FS_FCloseFile(f);		//close the file

	G_Printf("Successfully saved %s\n", filename);	//report success
}


/*///////////////////////////////////////////////////
BotFindLongTermGoal
Searches through all the nodes on a map and finds a future
goal for the bot to work toward.  Not very great at the moment
and seems to have major issues with axis on beach
TODO: take objective priority into account
*////////////////////////////////////////////////////
void BotFindLongTermGoal(gentity_t *bot)
{
	int i;
	int j = 0;
	int objNums[10];
	int dist;
	int node = NODE_INVALID;
	int bestdist = -1;
	int bestnode = NODE_INVALID;

	for (i = 0; i < numNodes; i++)			//loop through all the nodes in the map
	{
		if (nodes[i].flags == NODE_INVALID)		//skip invalid nodes
			continue;
		//also skip any nodes that are unreachable
		if ((nodes[i].flags & NODE_ALLY_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_BLUE))
			continue;
		if ((nodes[i].flags & NODE_AXIS_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_RED))
			continue;
		if (i == botinfo[bot->s.number].currentNode)		//don't pick the node he's currently at
			continue;

		//if the bot is carrying an objective
		if ((bot->client->ps.powerups[PW_REDFLAG]) || (bot->client->ps.powerups[PW_BLUEFLAG]))
		{
			//set the bot's longterm goal to the deliver point of the objective for his team
			if ((nodes[i].flags & NODE_AXIS_DELIVER) && (bot->client->sess.sessionTeam == TEAM_RED))
			{
				botinfo[bot->s.number].longTermGoal = nodes[i].num;
				return;
			}
			if ((nodes[i].flags & NODE_ALLY_DELIVER) && (bot->client->sess.sessionTeam == TEAM_BLUE))
			{
				botinfo[bot->s.number].longTermGoal = nodes[i].num;
				return;
			}

			continue;	//force the flag-carrying bot to ignore all other objectives
		}

		if (nodes[i].flags & NODE_OBJECTIVE)	//if it's an objective node (currently all that is checked for)
		{
			//don't have non-engs look for dyno spots
			if ((nodes[i].objType == OBJECTIVE_DYNAMITE) && (bot->client->ps.stats[STAT_PLAYER_CLASS] != PC_ENGINEER))
				continue;

			//if it's a steal objective
			if ((nodes[i].objType == OBJECTIVE_STEAL) && (bot->client->sess.sessionTeam == nodes[i].objTeam))
			{
				//this is a virtual copy of the flag code below
				//if there's no linked objective, find one and link it so we don't have to keep searching
				if ((nodes[i].objEntity == ENTITYNUM_NONE) || (nodes[i].objEntity < MAX_CLIENTS))
				{
					gentity_t *obj;
					vec3_t range, mins, maxs;
					int numTouch, k;
					int touch[40];

					VectorSet(range, 300, 300, 20);
					VectorSubtract(nodes[i].pos, range, mins);
					VectorAdd(nodes[i].pos, range, maxs);

					numTouch = trap_EntitiesInBox(mins, maxs, touch, 40);

					for (k = 0; k < numTouch; k++)
					{
						obj = &(g_entities[touch[k]]);

						//I *think* these are the only classnames the objectives will ever have
						if ((!Q_stricmp(obj->classname, "team_CTF_blueflag")) || (!Q_stricmp(obj->classname, "team_CTF_redflag")))
							break;
					}

					//we didn't find an objective near the node...someone picked it up?
					if (k == numTouch)
						continue;

					nodes[i].objEntity = obj->s.number;
				}
			}

			if (nodes[i].objType == OBJECTIVE_CAPTURE)		//if it's a flag
			{
				//if the actual flag isn't linked to the node, find it and link it, so future searches will be faster
				if ((nodes[i].objEntity == ENTITYNUM_NONE) || (nodes[i].objEntity < MAX_CLIENTS))
				{
					gentity_t *flag;
					vec3_t range, mins, maxs;
					int numTouch, k;
					int touch[40];

					//basically copied from BotDoFlag
					VectorSet(range, 300, 300, 20);
					VectorSubtract(nodes[i].pos, range, mins);
					VectorAdd(nodes[i].pos, range, maxs);

					//get all the entities within that area
					numTouch = trap_EntitiesInBox(mins, maxs, touch, 40);

					for (k = 0; k < numTouch; k++)	//loop through all the found entities
					{
						flag = &(g_entities[touch[k]]);

						if (!Q_stricmp(flag->classname, "team_WOLF_checkpoint"))	//stop looping once we find the flag
							break;
					}

					//didn't find a flag here...this could be a major problem, but we'll ignore it here
					if (k == numTouch)
						continue;

					nodes[i].objEntity = flag->s.number;	//found a flag, so link it to the node for speed
				}
				
				//if the flag is an objective for this bot's team
				if (nodes[i].objTeam == bot->client->sess.sessionTeam)
				{
					//skip it if the bot's team already has the flag
					if (g_entities[nodes[i].objEntity].count == bot->client->sess.sessionTeam)
						continue;
				}
				else	//if the flag is an objective for the other team
				{
					//skip it if the other team does not have the flag (doing this also handles when no one has the flag)
					if (g_entities[nodes[i].objEntity].count != OtherTeam(bot->client->sess.sessionTeam))
						continue;
				}
			}

			objNums[j] = i;						//record the node number
			j++;								//keep a count of how many objectives we find
		}

		if (j == 10)							//if we find more than 10, stop (we only need to pick one, and 10 is more than enough selection)
			break;
	}

	//get a valid value for a starting node, based on the bot's position
	node = botinfo[bot->s.number].currentNode;
	if (node == -1)
		node = BotFindClosestNode(bot);
	if (node == -1)							//if we can't get a valid value, something's wrong, so return
		return;

	for (i = 0; i < j; i++)					//loop through all the objective nodes we found
	{
		//BOTS TW - I *HATE* GetCost...so use linear distance for now
		//dist = GetCost(node, objNums[i], 256);		//get the approximate distance to the node
		dist = VectorDistance(nodes[node].pos, nodes[objNums[i]].pos);
		/*if (dist == -1)									//if the function couldn't determine the distance
		{
			if (bestdist == -1)
				bestnode = objNums[i];					//if we haven't selected any nodes yet, make sure to pick this one in case it's the only one

			continue;									//go to the next node since we don't have a distance
		}*/

		dist *= (1 + (crandom()/2));					//give small fluctuations to the distance so they don't always pick the same obj
		if ((dist < bestdist) || (bestdist == -1))		//if this node is the closest we've found so far
		{
			bestdist = dist;
			bestnode = objNums[i];						//mark it as the best goal
		}
	}

	botinfo[bot->s.number].longTermGoal = bestnode;		//set the bot's long-term goal to the best node we found
}

/*////////////////////////////////////////////
UpdateNodes
Called every frame to update the unreachable/objective flags for nodes,
so that the bots are aware an obstruction has been removed, etc
*/////////////////////////////////////////////
void UpdateNodes()
{
	int i, j;
	qboolean open[3];

	for (i = 0; i < numNodes; i++)		//loop through all the nodes
	{
		//if this node has the unreachable flag set
		if (nodes[i].flags & (NODE_ALLY_UNREACHABLE|NODE_AXIS_UNREACHABLE))
		{
			int j;

			for (j = 0; j < 3; j++)		//loop through all the entity nums linked to the unreachable flag
			{
				open[j] = qfalse;		//default to false (still blocked)

				//if there isn't a linked object, skip it
				if ((nodes[i].objectNum[j] == ENTITYNUM_NONE) || (nodes[i].objectNum[j] == 0))
					continue;

				if (g_entities[nodes[i].objectNum[j]].spawnflags == 0)	//if the spawnflags are 0, it is now open
					open[j] = qtrue;
			}

			if (open[0] || open[1] || open[2])	//if any of them are now open
				nodes[i].flags &= ~(NODE_ALLY_UNREACHABLE|NODE_AXIS_UNREACHABLE);	//clear the unreachable flags (and the objective flag, so bots don't keep setting it as a goal)
		}

		//if we blew up a dynamite objective, then clear the objective flag
		if ((nodes[i].flags & NODE_OBJECTIVE) && (nodes[i].objType == OBJECTIVE_DYNAMITE))
		{
			if (g_entities[nodes[i].objEntity].spawnflags == 0)
				nodes[i].flags &= ~NODE_OBJECTIVE;
		}

		//loop through all the links from this node
		for (j = 0; j < nodes[i].numLinks; j++)
		{
			if (nodes[i].links[j].flags & PATH_BLOCKED)	//if the path is blocked
			{
				trace_t tr;
				vec3_t start, end;

				VectorCopy(nodes[i].pos, start);
				VectorCopy(nodes[nodes[i].links[j].target].pos, end);
				
				//do a trace to see if there is still anything blocking the path
				trap_Trace(&tr, start, NULL, NULL, end, ENTITYNUM_NONE, MASK_DEADSOLID);

				//if the path is open, clear the 'blocked' flag
				if (tr.fraction == 1.0)
					nodes[i].links[j].flags &= ~PATH_BLOCKED;
			}
		}
	}
}
