#include "g_local.h"
#include "BotMain.h"

/***************************************************************************************
original code: TeamWhore

filename: BotState.c

Description: This file contains the functions to handle the various bot states

###################### Function Descriptions ##################################

void BotStateActions(gentity_t *bot, int *state)
	Provides high-level control of the different bot states, basically just calls the
	different state functions

void HandleDeathState(gentity_t *bot, int *state)
	Performs all the things that are done when the bot dies

void HandleAttackState(gentity_t *bot, int *state)
	Handles the attack state for the bot

void HandleWanderState(gentity_t *bot, int *state)
	This handles the 'wander' state - right now it doesn't do anyting, but it's pretty
	much just going to implement non-goal-seeking movement for a limited time

void HandleFollowState(gentity_t *bot, int *state)
	Handles when the bot is following another player

void HandleGiveItemState(gentity_t *bot, int *state)
	Called when the bot is trying to give an item (ammo/health) to himself or another player

void HandleReviveState(gentity_t *bot, int *state)
	Handles when a medic bot is trying to revive a wounded player

void HandleMoveState(gentity_t *bot, int *state)
	This function contains the code to implement bot movement through the nodes.  Bots will
	probably spend at least half of their time in this state on many maps

void HandleIdleState(gentity_t *bot, int *state)
	Handles BOTSTATE_STAND, which is pretty much only set when the bot is not currently doing
	anything.  At the moment, this state is only really set during botstate transitions, so that
	the bots don't react instantly to everything, but have at least a one frame delay

qboolean PerformNodeActions(gentity_t *bot)
	Called when the bot has arrived at a node.  Checks to see what kind of node this is
	and handles performing actions such as camping, dynamiting, using an mg42, etc

qboolean BotDoDynamite(gentity_t *bot)
	Called from an objective node with the OBJECTIVE_DYNAMITE flag set so that the bot will plant
	dynamite here if the objective has not been blown.  Currently does not check to see if the bot
	or any other bot has yet planted dynamite here.  Also does not currently have disarming code.

qboolean BotDoStealObjective(gentity_t *bot)
	Pretty obvious what this will do from the name.  Currently unimplemented :P

qboolean BotDoFlag(gentity_t *bot)
	Called from an objective node with the OBJECTIVE_CAPTURE flag set so that the bot will attempt
	to get the flag, provided his team does not already have it.  This code is currently functional,
	but not nearly as accurate as it should be.

###############################################################################

-----changes:---------------------
  March 4-5, 2004 
	TeamWhore - created this file and all it's contents :P
	    moved death code from Botthink to HandleDeathState
		moved stand/move code from Botthink to HandleNormalState
	    moved HandleAttackState here
		bots now don't tap out if a medic is in the area
		bots will call for a medic periodically when they are dead
		added code to HandleNormalState to allow for when bots are following a player, or trying
		    to give a player health/ammo
  March 5, 2004
    TeamWhore - changed enemy seeking in HandleAttackState to use botinfo.seekTime + SEEK_TIMEOUT
	    changed calling for medic code to use botinfo.lastMedicChatTime + CHAT_REPEATTIME
		started adding node movement code to HandleNormalState
  March 6, 2004
	TeamWhore - added better node handling to HandleNormalState, still far from finished
		added PerformNodeActions function to handle situations when the bot arrives at a node - unfinished
		added BotDoDynamite, BotDoStealObjective and BotDoFlag functions to handles these situations - unfinished
  March 6-7, 2004
	TeamWhore - added new function BotStateActions to act as a controller and call the appropriate functions
			based on a bot's state
		added HandleWanderState function to accomodate new wander state
		added HandleFollowState function to handle when a bot is following another player
		added HandleGiveItemState function control when a bot wants to give an item to himself or another player
		added HandleReviveState function for when medics are trying to revive a wounded teammate
		added HandleIdleState to handle BOTSTATE_STAND when the bot isn't doing anything in particular
		added HandleMoveState to handle a bot's movement using nodes
		moved movement code from HandleNormalState to HandleMovementState
		deleted obsolete function HandleNormalState
  March 9, 2004
	TeamWhore - changed node movement code to always call PerformNodeActions upon reaching a node
		added code to BotDoDynamite so that engineers who have already planted at a dynamite objective node will
			not try to plant again
		implemented BotDoFlag
		changed HandleIdleState to clear a bot's current/next nodes, in case he had moved away from them
  March 11, 2004
	TeamWhore - fixed bot's death animation so that they no longer crumple strangely
		changed BotStateActions to call LocateEnemy even in BOTSTATE_ATTACK
		added check to BotStateActions to prevent insta-gibbed bots from calling for a medic
  April 3, 2004
	TeamWhore - moved a lot of death/spawn code from ClientSpawn to HandleDeathState
		removed part of the medic chat 'fix' from last update - didn't work
		bots theoretically now don't call for a med when they die unless a med is nearby
		added code for HandleWanderState to allow for non-goal-based movement
		changed HandleMoveState to incorporate new pathfinding
		edited HandleIdleState to reset some values and to incorporate new pathfinding
		minor updates to BotDoFlag, still not great
		moved some functions into new file BotUtils.c
  April-July, 2004
	TeamWhore - played around a lot with the node movement code - seems to be almost semi-decent now
		bots now will move along node paths according to the path movement flags for each node (crouch, etc)
		made wandering behave the same as goal-based movement, but with no goal - at least for now
		added code to make the bots suicide if they are unable to move for a certain amount of time
		updated BotDoFlag so that it will correctly handle more flags, and not cause crashes/weird bot movement


****************************************************************************************/
extern void limbo (gentity_t *ent, qboolean makeCorpse);
extern void Cmd_Kill_f(gentity_t *ent);

/*////////////////////////////////////////////////////////
BotStateActions
Just provides high level control of the bot relating to his current state
just passes control to the specific state functions
*/////////////////////////////////////////////////////////
void BotStateActions(gentity_t *bot, int *state)
{
	qboolean spawnDelay = qfalse;

	if (level.time < bot->client->respawnTime + 1000)
		spawnDelay = qtrue;

	//if the bot was dead, but is now alive (respawn/revive)
	if ((*state == BOTSTATE_DEAD) && (bot->client->ps.stats[STAT_HEALTH] > 0))
		*state = BOTSTATE_STAND;		//reset the botstate since he's not dead

	if (bot->client->ps.pm_type == PM_DEAD)		//if the bot died
		HandleDeathState(bot, state);			//this function deals with everything when the bot is dead

	if (spawnDelay)
		return;								//handle everything pertaining to spawning etc, but nothing else

	if (bot->client->ps.powerups[PW_REDFLAG] || bot->client->ps.powerups[PW_BLUEFLAG])
		*state = BOTSTATE_CARRY;

	if (*state == BOTSTATE_CARRY)
		HandleCarryObjectiveState(bot, state);

	if (*state == BOTSTATE_REVIVE)			//reviving takes precedence over everything
		HandleReviveState(bot, state);

	if (*state == BOTSTATE_FOLLOW)
		HandleFollowState(bot, state);

	//second half of this statement is here because we allow bot to give himself ammo without changing the state
	if ((*state == BOTSTATE_GIVEITEM) || (botinfo[bot->s.number].giveItemTarget == bot->s.number))
		HandleGiveItemState(bot, state);

	//if the bot is attacking and not trying to give himself ammo
	if ((*state == BOTSTATE_ATTACK) && (botinfo[bot->s.number].givingSelfItemState == 0))
		HandleAttackState(bot, state);

	if (*state == BOTSTATE_MOVE)
		HandleMoveState(bot, state);

	if (*state == BOTSTATE_WANDER)
		HandleWanderState(bot, state);

	if ((*state != BOTSTATE_CARRY) && (*state != BOTSTATE_REVIVE) && (*state != BOTSTATE_GIVEITEM))
	{
		BotCheckAmmoState(bot, state);

		if (LocateEnemy(bot))			//do this even if we're attacking, in case we find a better enemy
			*state = BOTSTATE_ATTACK;	//'better' should be defined somehow in LocateEnemy
	}

	if (*state == BOTSTATE_STAND)
		HandleIdleState(bot, state);
}

/*//////////////////////////////////////////////////////
HandleDeathState
this is called from Botthink, and handles what happens if the bot is dead/wounded
Essentially, this function resets whatever needs to be, and handles things like calling
for a medic and tapping out
*///////////////////////////////////////////////////////
void HandleDeathState(gentity_t *bot, int *state)
{
	int respawnTimer;

	bot->client->buttons = 0;				//reset a whole bunch of stuff
	bot->client->wbuttons = 0;
	*state = BOTSTATE_DEAD;
	bot->enemy = NULL;
	bot->client->ps.persistant[PERS_ATTACKER] = -1;

	//reset a lot of the botinfo vars, moved here from g_client.c
	botinfo[bot->s.number].giveItemTarget = ENTITYNUM_NONE;
	botinfo[bot->s.number].followingNum = ENTITYNUM_NONE;
	botinfo[bot->s.number].followingTime = 0;
	botinfo[bot->s.number].givingSelfItemState = 0;
	botinfo[bot->s.number].seekTime = 0;
	botinfo[bot->s.number].seekDir = DIR_NOMOVE;
	botinfo[bot->s.number].numFramesIdle = 0;
	botinfo[bot->s.number].reviveTarget = ENTITYNUM_NONE;
	botinfo[bot->s.number].framesPassed = 0;
	botinfo[bot->s.number].wanderTime = 0;
	botinfo[bot->s.number].pathsize = 0;	//if this is reset, don't have to bother with pathlist
	botinfo[bot->s.number].moveFlags = PATH_NORMAL;
	botinfo[bot->s.number].suicideTime = 0;
	botinfo[bot->s.number].onNode = qfalse;

	//this is necessary to prevent the death animation from being all screwy
	bot->s.angles[0] = 0;
	bot->s.angles[1] = bot->client->ps.viewangles[1];
	bot->s.angles[2] = 0;
	VectorCopy(bot->s.angles, bot->client->ps.viewangles);

	if (!(bot->client->ps.pm_flags & PMF_LIMBO))	//don't try to tap out if we're already gibbed
	{
		//force respawn if timer is up
		if (bot->client->sess.sessionTeam == TEAM_BLUE)
			respawnTimer = g_bluelimbotime.integer - (level.time % g_bluelimbotime.integer);	//in milliseconds
		else
			respawnTimer = g_redlimbotime.integer - (level.time % g_redlimbotime.integer);

		//if there are less than 2 seconds till respawn, and there is not a med in the vicinity (see WolfFindMedic)
		if ((respawnTimer <= 2000) && (bot->client->ps.viewlocked != 7))	
		{
			limbo(bot, qtrue);			//tap out
			return;						//don't call for a med if he tapped
		}

		if (bot->client->ps.viewlocked == 7)	//if there is a medic in the vicinity - see WolfFindMedic
		{
			if (random() > 0.84)		//space it out a little bit, so timing isn't EXACTLY every CHAT_REPEATTIME milliseconds...
			{
				BotTalk(bot, "medic", qtrue, qtrue);
			}
		}
	}
}


/*//////////////////////////////////////
HandleAttackState
this code is just here so Botthink isn't totally cluttered
seeking is simplified, just crouches and strafes left or right, looking for enemy again
keeps eyes focused on enemy pos but won't shoot unless he's visible
if bot runs into an obstacle seeking, he reverses direction
/*//////////////////////////////////////
void HandleAttackState(gentity_t *bot, int *state)
{
	//Check if the enemy is dead, so stop attacking by setting bot state to BOTSTATE_STAND
	if (bot->enemy->client->ps.stats[STAT_HEALTH] <= 0)
	{
		*state = BOTSTATE_STAND;
		bot->enemy = NULL;
		bot->client->ps.persistant[PERS_ATTACKER] = -1;		//reset enemy info
		ChooseWeapon(bot, 0);					//so switch after shooting panz, etc
	}

	//if the enemy is (still) alive
	if (bot->enemy)
	{
		// Check if we can see the enemy from our current position
		if (visible(bot, bot->enemy))
		{
			//if we can see him, shoot him!
			AttackEnemy(bot, state);
		}
		else
		{
			// Before the bot could see the enemy, but now it can't (example: if the enemy moved behind an obstacle)
			if (!botinfo[bot->s.number].seekTime)
			{
				botinfo[bot->s.number].seekTime = level.time;		//set the time we started seeking

				//fix all this to be smarter about direction choosing...based on dist etc? -TW
				// es_fix : maybe we could take the enemy's last know coordinates and find the 
				//			nearest node to that position, from there, that would then be this bots
				//			current goal. Also if the bots health is too low from the initial encounter
				//			with the enemy, seek the nearest medic who isn't in combat (snap medic's
				//          position to nearest waypoint).
				//TW - yeah, but not relative to nodes, because bot might not be near nodes (it CAN happen...)

				//randomly select a direction to strafe
				if (random() > 0.5)
					botinfo[bot->s.number].seekDir = DIR_LEFT;		//strafe left
				else
					botinfo[bot->s.number].seekDir = DIR_RIGHT;		//strafe right

				//turn towards the enemy
				//TW - this should eventually be changed to 'turn towards the enemy's last known position'
				FaceEnemy(bot);
			}
			//If the bot is still seeking
			else if (level.time < (botinfo[bot->s.number].seekTime + SEEK_TIMEOUT))
			{
				//turn towards the enemy again...see above note
				FaceEnemy(bot);

				//check if the bot can keep moving in the same direction
				//if he can, strafe that direction (and crouch-walk)
				if (CanMove(bot, botinfo[bot->s.number].seekDir))
					BotMove(bot, botinfo[bot->s.number].seekDir, BOTMOVE_CROUCH);
				else
				{
					//don't move anywhere, but stay crouching
					BotMove(bot, DIR_NOMOVE, BOTMOVE_CROUCH);

					//reverse direction
					if (botinfo[bot->s.number].seekDir == DIR_LEFT)
						botinfo[bot->s.number].seekDir = DIR_RIGHT;
					else
						botinfo[bot->s.number].seekDir = DIR_LEFT;
				}
			}
			else
			{
				//the bot can't find enemy, so reset everything
				botinfo[bot->s.number].seekTime = 0;
				botinfo[bot->s.number].seekDir = DIR_NOMOVE;
				bot->enemy = NULL;
				bot->client->ps.persistant[PERS_ATTACKER] = -1;
				*state = BOTSTATE_STAND;
			}
		}
	}
}

/*//////////////////////////////////////////////
HandleWanderState
Controller for BOTSTATE_WANDER
Basically copies HandleMoveState, minus the longterm goal bits
The 'next node' code is also different in BotGetNextNode
Really just supposed to have the bot run around with no goal
(but hopefully not in circles) for a little while
*///////////////////////////////////////////////
void HandleWanderState(gentity_t *bot, int *state)
{
	//if the bot has wandered for long enough
	if (level.time >= (botinfo[bot->s.number].wanderTime + WANDER_TIMEOUT))
	{
		*state = BOTSTATE_STAND;	//reset the state and return
		return;
	}

	if (botinfo[bot->s.number].nextNode == NODE_INVALID)					//if the bot doesn't have a target
		botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);		//get one!

	if (botinfo[bot->s.number].nextNode != NODE_INVALID)	//check again because BotGetNextNode could have returned NDOE_INVALID
	{
		//this code is copied straight from HandleMoveState...obviously a temporary 'fix' - this will be updated later
		vec3_t v;
		float dist;

		//get the direction and distance to the next node
		VectorSubtract(nodes[botinfo[bot->s.number].nextNode].pos, bot->r.currentOrigin, v);
		dist = VectorLength(v);
		vectoangles(v, bot->s.angles);	//turn to face it

		//if the bot isn't at this node yet and he isn't moving at a node
		if ((dist > 32) && (botinfo[bot->s.number].movingAtNode == qfalse))
		{
			//copy over the movement flags for this link, so the bot knows how to move
			if (botinfo[bot->s.number].currentNode != NODE_INVALID)		//invalid if they just started moving
			{
				int i;
				node_t *cur = &(nodes[botinfo[bot->s.number].currentNode]);

				//loop through the links until we find the correct one
				for (i = 0; i < MAX_LINKS; i++)
				{
					if (cur->links[i].target == botinfo[bot->s.number].nextNode)
					{
						//set the flags and break, since we found the flags
						botinfo[bot->s.number].moveFlags = cur->links[i].flags;
						break;
					}
				}
			}

			//check if he can move towards it
			if (CanMove(bot, DIR_FORWARD))
			{
				BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);	//if so, then move!
				botinfo[bot->s.number].onNode = qfalse;
			}
			else
			{
				//first try to move closer to the current node and see if that helps
				if (!BotMoveNearerNode(bot))
				{
					//if not, see if different types of movement will work
					if (!BotSpecialMove(bot))
					{
						//at this point, nothing appears to work, so try brute force and just go forward :)
						BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);
					}
				}
			}
		}
		else	//the bot has reached this node
		{
			//preform actions at this node, if any
			if (PerformNodeActions(bot))	//returns qtrue when the bot is done at this node
			{
				//once we're done at this node, move all the nodes up one, get a new nextNode, and reset movingAtNode and framesAtNode
				botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode;
				botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode;
				botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);
				botinfo[bot->s.number].movingAtNode = qfalse;
				botinfo[bot->s.number].framesAtNode = 0;
				botinfo[bot->s.number].onNode = qfalse;
			}
		}
	}

	//if the bot really isn't moving (and should be), and it's gone on for a while, have him suicide
	if ((botinfo[bot->s.number].movingAtNode == qfalse) && (VectorLength(bot->client->ps.velocity) < 32))
	{
		if ((botinfo[bot->s.number].suicideTime) && (level.time >= botinfo[bot->s.number].suicideTime))
			Cmd_Kill_f(bot);
	}
	else
	{
		//when he moves 'normally', make sure to update this so they won't suicide unnecessarily
		botinfo[bot->s.number].suicideTime = level.time + SUICIDE_TIMEOUT;
	}
}

/*//////////////////////////////////////////////
HandleFollowState
checks to make sure the bot is still supposed to be
following someone and does this or resets the follow
state
*///////////////////////////////////////////////
void HandleFollowState(gentity_t *bot, int *state)
{
	if (botinfo[bot->s.number].followingNum == ENTITYNUM_NONE)		//if the bot isn't following anyone
	{
		*state = BOTSTATE_STAND;		//reset the state and return
		return;
	}

	if (level.time > (botinfo[bot->s.number].followingTime + FOLLOW_TIMEOUT))	//follow them for FOLLOW_TIMEOUT milliseconds
	{
		botinfo[bot->s.number].followingNum = ENTITYNUM_NONE;	//if we're done following them, reset everything and return
		botinfo[bot->s.number].followingTime = 0;
		*state = BOTSTATE_STAND;
		return;
	}
	
	BotFollowPlayer(bot);		//perform the actual following functionality
}

/*///////////////////////////////////////////////
HandleGiveItemState
Called when a bot wants to give himself or another player an item
some code moved here from BotMisc.c
*////////////////////////////////////////////////
void HandleGiveItemState(gentity_t *bot, int *state)
{
	if (botinfo[bot->s.number].giveItemTarget == ENTITYNUM_NONE)	//if the bot isn't trying to give anyone an item
	{
		*state = BOTSTATE_STAND;			//reset the state info and return
		return;
	}

	if (botinfo[bot->s.number].giveItemTarget == bot->s.number)	//if he's giving himself an item
	{
		//if the bot isn't already giving himself ammo/health
		if (botinfo[bot->s.number].givingSelfItemState == 0)
		{
			int dir;
			dir = BotFindFreeDirection(bot);	//find an open direction

			switch(dir)							//this block turns the bot towards whichever way was open
			{
			case DIR_FORWARD:
				break;
			case DIR_RIGHT:
				bot->s.angles[1] -= 90;
				break;
			case DIR_LEFT:
				bot->s.angles[1] += 90;
				break;
			case DIR_BACKWARD:
				bot->s.angles[1] -= 180;
				break;
			default:								//DIR_NOMOVE - we aren't clear in any direction
				break;
			}

			if (dir != DIR_NOMOVE)		//if the bot can move in some direction
			{
				if (bot->client->ps.stats[STAT_PLAYER_CLASS] == PC_LT)
					ChooseWeapon(bot, WP_AMMO);							//select ammo bag
				else
					ChooseWeapon(bot, WP_MEDKIT);						//select medpack

				botinfo[bot->s.number].givingSelfItemState = 1;		//found a direction
			}
		}
		//if the bot is facing an open direction but hasn't thrown the pack yet
		else if (botinfo[bot->s.number].givingSelfItemState < 4)	//4 because it takes a few frames for the bot to throw the bag, but only if he has a thompson, lol
		{
			botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;		//throw it!
			botinfo[bot->s.number].givingSelfItemState++;				//increment this so we know another frame has passed
		}
		//if the bot has thrown a bag but hasn't picked it up yet
		else
		{
			//move toward the bag, don't reset givingSelfItemState until he's picked it up
			BotMove(bot, DIR_FORWARD, 0);
		}
	}
	else
	{
		if (bot->client->ps.stats[STAT_PLAYER_CLASS] == PC_LT)	//it the bot is a lt, select an ammo pack
			ChooseWeapon(bot, WP_AMMO);
		else
			ChooseWeapon(bot, WP_MEDKIT);						//otherwise he's a med, so select a medpack

		BotGiveItemToPlayer(bot, state);				//keep trying to give the item to the person
	}
}

/*////////////////////////////////////////////
HandleReviveState
Controller for BOTSTATE_REVIVE, basically just calls
BotRevivePlayer
*/////////////////////////////////////////////
void HandleReviveState(gentity_t *bot, int *state)
{
	if ((botinfo[bot->s.number].reviveTarget == ENTITYNUM_NONE) || (g_entities[botinfo[bot->s.number].reviveTarget].client->ps.pm_type != PM_DEAD))
	{
		botinfo[bot->s.number].reviveTarget = ENTITYNUM_NONE;
		*state = BOTSTATE_STAND;
		return;
	}

	BotRevivePlayer(bot);
}

/*/////////////////////////////////////////////
HandleMoveState
Called to handle movement of the bot between nodes
Code was originally in HandleNormalState, was moved here
*//////////////////////////////////////////////
void HandleMoveState(gentity_t *bot, int *state)
{
	//if the bot has a goal
	if (botinfo[bot->s.number].longTermGoal != NODE_INVALID)
	{
		//if he doesn't have a node to move towards
		if (botinfo[bot->s.number].nextNode == NODE_INVALID)
		{
			//if he has a path already, just set nextNode to the next node on his path
			if (botinfo[bot->s.number].pathsize > 0)
				botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);
			//otherwise, the path is invalid, so invalidate everything and wander
			else
			{
				botinfo[bot->s.number].longTermGoal = NODE_INVALID;
				*state = BOTSTATE_WANDER;
				botinfo[bot->s.number].wanderTime = level.time;
				//reset all the path/node info for movement
				//TODO: create a state-switching function, and throw this code in it as well
				botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode = NODE_INVALID;
			}
		}
		else	//we have a node to move to
		{
			vec3_t v;
			float dist;

			//get the direction and distance to the next node
			VectorSubtract(nodes[botinfo[bot->s.number].nextNode].pos, bot->r.currentOrigin, v);
			vectoangles(v, bot->s.angles);	//turn to face it

			//TODO: later, maybe check for ladders, and don't drop vertical component when moving on one
			dist = VectorLength(v);

			//if the bot isn't at this node yet and he isn't moving at a node
			if ((dist > 32) && (botinfo[bot->s.number].movingAtNode == qfalse))
			{
				//copy over the movement flags for this link, so the bot knows how to move
				if (botinfo[bot->s.number].currentNode != NODE_INVALID)		//invalid if they just started moving
				{
					int i;
					node_t *cur = &(nodes[botinfo[bot->s.number].currentNode]);

					//loop through the links until we find the correct one
					for (i = 0; i < MAX_LINKS; i++)
					{
						if (cur->links[i].target == botinfo[bot->s.number].nextNode)
						{
							//set the flags and break, since we found what we wanted
							botinfo[bot->s.number].moveFlags = cur->links[i].flags;
							break;
						}
					}
				}

				//check if he can move towards it
				if (CanMove(bot, DIR_FORWARD))
				{
					BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);	//if so, then move!
					botinfo[bot->s.number].onNode = qfalse;
				}
				else
				{
					//first try to move closer to the current node and see if that helps
					if (!BotMoveNearerNode(bot))
					{
						//if not, see if different types of movement will work
						if (!BotSpecialMove(bot))
						{
							//at this point, nothing appears to work, so try brute force and just go forward
							BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);
						}
					}
				}
			}
			else	//the bot has reached this node
			{
				//preform actions at this node, if any
				if (PerformNodeActions(bot))	//returns qtrue when the bot is done at this node
				{
					//once we're done at this node, move all the nodes up one, get a new nextNode, and reset movingAtNode and framesAtNode
					botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode;
					botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode;
					botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);
					botinfo[bot->s.number].movingAtNode = qfalse;
					botinfo[bot->s.number].framesAtNode = 0;
					botinfo[bot->s.number].onNode = qfalse;

					//if the bot just reached his goal, reset the goal and wander for a bit
					if (botinfo[bot->s.number].currentNode == botinfo[bot->s.number].longTermGoal)
					{
						botinfo[bot->s.number].longTermGoal = NODE_INVALID;
						*state = BOTSTATE_WANDER;
						botinfo[bot->s.number].wanderTime = level.time;
						botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode = NODE_INVALID;
					}
				}
			}
		} //(in)valid node
	} //has a goal
	else
	{
		//if the bot has no goal, just wander
		*state = BOTSTATE_WANDER;
		botinfo[bot->s.number].wanderTime = level.time;
		botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode = NODE_INVALID;
	}

	//if the bot really isn't moving (and should be), and it's gone on for a while, have him suicide
	if ((botinfo[bot->s.number].movingAtNode == qfalse) && (VectorLength(bot->client->ps.velocity) < 32))
	{
		if ((botinfo[bot->s.number].suicideTime) && (level.time >= botinfo[bot->s.number].suicideTime))
			Cmd_Kill_f(bot);
	}
	else
	{
		//when he moves 'normally', make sure to update this so they won't suicide unnecessarily
		botinfo[bot->s.number].suicideTime = level.time + SUICIDE_TIMEOUT;
	}
}

/*////////////////////////////////////////////
HandleIdleState
Called when the bot is in BOTSTATE_STAND
Basically this function just provides a one frame
lag for state changes
Update - now it also sets up the bot's path for long-term goal
seeking
*/////////////////////////////////////////////
void HandleIdleState(gentity_t *bot, int *state)
{
	//if the bot doesn't have a goal, try to find one
	if (botinfo[bot->s.number].longTermGoal == NODE_INVALID)
		BotFindLongTermGoal(bot);

	botinfo[bot->s.number].previousNode = NODE_INVALID;		//by resetting these, the bot will look for the closest node again
	botinfo[bot->s.number].currentNode = NODE_INVALID;		//if we don't do this, but he moved far away from his current and/or next node
	botinfo[bot->s.number].nextNode = NODE_INVALID;			//then he will try to run straight at it, not following node paths, so he'll run into things

	botinfo[bot->s.number].suicideTime = 0;
	botinfo[bot->s.number].onNode = qfalse;

	//create/reset path for the bot to travel (it was either invalid, or he could have moved off course)
	if (botinfo[bot->s.number].longTermGoal != NODE_INVALID)
	{
		//this needs to be moved/staggered, so that every bot isn't calling it at the same time (lag potential)
		short int node = BotFindClosestNode(bot);		//we have to start with an actual node, which the bot likely isn't at, so pick the nearest one
		botinfo[bot->s.number].pathsize = CreatePathAStar(bot, node, botinfo[bot->s.number].longTermGoal, botinfo[bot->s.number].pathlist);
		botinfo[bot->s.number].nextNode = node;		//move to this node first, since it's where our path starts from
	}

	*state = BOTSTATE_MOVE;
}

/*////////////////////////////////////////////
HandleCarryObjectiveState
Called when the bot has the objective so that the bot will
try to get to the drop-off point without switching to any 
other state (this may be altered later for things like revives etc)
*/////////////////////////////////////////////
void HandleCarryObjectiveState(gentity_t *bot, int *state)
{
	qboolean search = qfalse;

	//if for some reason we're in this state without an objective...
	if (!(bot->client->ps.powerups[PW_REDFLAG]) && !(bot->client->ps.powerups[PW_BLUEFLAG]))
	{
		//reset the state and goal, and return
		*state = BOTSTATE_STAND;
		botinfo[bot->s.number].longTermGoal = NODE_INVALID;
		return;
	}

	//check if the bot has set the obj return point as his goal
	if (bot->client->sess.sessionTeam == TEAM_RED)
	{
		if (!(nodes[botinfo[bot->s.number].longTermGoal].flags & NODE_AXIS_DELIVER))
			search = qtrue;
	}
	else
	{
		if (!(nodes[botinfo[bot->s.number].longTermGoal].flags & NODE_ALLY_DELIVER))
			search = qtrue;
	}

	if (search)	//if not, set it up and create the bot's path to the return point
	{
		short int node = BotFindClosestNode(bot);
		BotFindLongTermGoal(bot);

		//make sure the pathlist is cleared before adding a new path to it
		memset(botinfo[bot->s.number].pathlist, NODE_INVALID, 128);
		botinfo[bot->s.number].pathsize = CreatePathAStar(bot, node, botinfo[bot->s.number].longTermGoal, botinfo[bot->s.number].pathlist);
		botinfo[bot->s.number].nextNode = node;
	}

	//shouldn't happen, but again, just to be safe...
	if (botinfo[bot->s.number].nextNode == NODE_INVALID)
		botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);

	//and even safer...
	if (botinfo[bot->s.number].nextNode != NODE_INVALID)
	{
		vec3_t v;
		float dist;

		//find the next node and face it
		VectorSubtract(nodes[botinfo[bot->s.number].nextNode].pos, bot->r.currentOrigin, v);
		dist = VectorLength(v);
		vectoangles(v, bot->s.angles);

		//if we haven't reached it yet
		if (dist > 32)
		{
			if (botinfo[bot->s.number].moveFlags == -1)		//means they haven't been set yet
			{
				int i;
				node_t *cur = &(nodes[botinfo[bot->s.number].currentNode]);

				//loop through the links until we find the correct one
				for (i = 0; i < MAX_LINKS; i++)
				{
					if (cur->links[i].target == botinfo[bot->s.number].nextNode)
					{
						//set the flags and break, since we found what we wanted
						botinfo[bot->s.number].moveFlags = cur->links[i].flags;
						break;
					}
				}
			}

			//sprint most of the time while carrying the objective (not ALL so maybe it will recharge...doubt it)
			if (random() > 0.2)
				botinfo[bot->s.number].moveFlags |= BOTMOVE_SPRINT;

			//try to move in all the old, fun, familiar ways
			if (CanMove(bot, DIR_FORWARD))
			{
				BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);
				botinfo[bot->s.number].onNode = qfalse;
			}
			else
			{
				if (!BotMoveNearerNode(bot))
				{
					if (!BotSpecialMove(bot))
						BotMove(bot, DIR_FORWARD, botinfo[bot->s.number].moveFlags);
				}
			}
		}
		else
		{
			//if we're at a node, update all the info and get the next node in the path
			botinfo[bot->s.number].previousNode = botinfo[bot->s.number].currentNode;
			botinfo[bot->s.number].currentNode = botinfo[bot->s.number].nextNode;
			botinfo[bot->s.number].nextNode = BotGetNextNode(bot, state);

			botinfo[bot->s.number].moveFlags = -1;	//reset this so we know to check the node data
			botinfo[bot->s.number].onNode = qfalse;
		}

		if (VectorLength(bot->client->ps.velocity) < 32)
		{
			if ((botinfo[bot->s.number].suicideTime) && (level.time >= botinfo[bot->s.number].suicideTime))
				Cmd_Kill_f(bot);
		}
		else
		{
			//when he moves 'normally', make sure to update this so they won't suicide unnecessarily
			//make it a little longer than for the other bots, since the obj is important
			botinfo[bot->s.number].suicideTime = level.time + (1.5 * SUICIDE_TIMEOUT);
		}
	}
}

/*////////////////////////////////////////////
PerformNodeActions
Called at nodes to offer high-level control of what actions 
to perform at any non-normal nodes.  Currently more
a skeleton function than anything useful
*/////////////////////////////////////////////
qboolean PerformNodeActions(gentity_t *bot)
{
	int node = botinfo[bot->s.number].nextNode;

	botinfo[bot->s.number].framesAtNode++;

	if (nodes[node].flags & NODE_OBJECTIVE)
	{
		int objType = nodes[node].objType;		//get the type of objective this is

		switch (objType)
		{
		case OBJECTIVE_DYNAMITE:
			return BotDoDynamite(bot);
		case OBJECTIVE_STEAL:
			return BotDoStealObjective(bot);
		case OBJECTIVE_CAPTURE:
			return BotDoFlag(bot);
		}
	}
	else if (nodes[node].flags & NODE_ACTIVATE)
	{}
	else if (nodes[node].flags & NODE_CAMP)
	{}
	else if (nodes[node].flags & NODE_MG42)
	{}

	return qtrue;	//if it's a normal node
}

/*///////////////////////////////////////////
BotDoDynamite
Called when a bot is at a dynamite objective node
Handles planting dynamite, still unfinished
TODO: don't plant if armed dyno is already present
*////////////////////////////////////////////
qboolean BotDoDynamite(gentity_t *bot)
{
	gentity_t *ent, *dyno;
	vec3_t pos;
	int dir = DIR_NOMOVE;

	if (bot->client->ps.stats[STAT_PLAYER_CLASS] != PC_ENGINEER)
		return qtrue;											//FIXME: change to wandering? (probably just ignore, since this shouldn't be their goal anyway)

	ent = &(g_entities[nodes[botinfo[bot->s.number].nextNode].objEntity]);

	if (ent->spawnflags == 0)								//already been blown
		return qtrue;

	dyno = &(g_entities[bot->grenadeFired]);

	if (Q_stricmp(dyno->classname, "dynamite") != 0)			//if the dyno doesn't exist yet
	{
		if (level.time - bot->client->ps.classWeaponTime < g_engineerChargeTime.integer)	//doesn't have charge
			return qtrue;

		VectorAdd(ent->r.mins, ent->r.maxs, pos);
		VectorScale(pos, 0.5, pos);
		pos[2] = ent->r.mins[2];								//face the bottom center of target
	}
	else
		VectorCopy(dyno->r.currentOrigin, pos);					//face dyno

	VectorSubtract(pos, bot->r.currentOrigin, pos);
	vectoangles(pos, bot->s.angles);

	if ((!Q_stricmp(dyno->classname, "dynamite")) && (VectorLength(pos) > 60))	//too far away from dyno to arm
	{
		dir = DIR_FORWARD;
		botinfo[bot->s.number].movingAtNode = qtrue;
	}

	BotMove(bot, dir, BOTMOVE_CROUCH);	//crouch to arm dyno and/or move towards it

	if (botinfo[bot->s.number].framesAtNode < 3)
		return qfalse;			//wait a few frames once we reach node
	else if (botinfo[bot->s.number].framesAtNode == 3)
		ChooseWeapon(bot, WP_DYNAMITE);		//select the dyno
	else if (botinfo[bot->s.number].framesAtNode == 8)
		botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;	//throw the dyno
	else if (botinfo[bot->s.number].framesAtNode == 15)
		ChooseWeapon(bot, WP_PLIERS);		//select the pliers
	else if (botinfo[bot->s.number].framesAtNode >= 20)
	{
		if (dyno->s.teamNum >= 4)					//not armed - see Weapon_Engineer
			botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;	//arm it
		else
		{
			ChooseWeapon(bot, 0);
			bot->grenadeFired = ENTITYNUM_NONE;		//clear this so it doesn't interfere with other bomb nodes
			return qtrue;
		}
	}

	return qfalse;
}

/*//////////////////////////////////////////
BotDoStealObjective
Called at a steal objective node. Searches the
area around the bot for an objective, and attempts
to have the bot take it if there is one present.
Currently much like BotDoFlag
*///////////////////////////////////////////
qboolean BotDoStealObjective(gentity_t *bot)
{
	vec3_t pos, mins, maxs, range, origin;
	int touch[MAX_GENTITIES], numTouch, i;
	gentity_t *obj;

	//if we already have an obj, return (this shouldn't happen, but it's safer)
	if (bot->client->ps.powerups[PW_REDFLAG] || bot->client->ps.powerups[PW_BLUEFLAG])
		return qtrue;

	if ((nodes[botinfo[bot->s.number].nextNode].objEntity != ENTITYNUM_NONE) && (nodes[botinfo[bot->s.number].nextNode].objEntity > MAX_CLIENTS))
		obj = &(g_entities[nodes[botinfo[bot->s.number].nextNode].objEntity]);
	else
	{
		//if there isn't a linked obj with this node, search for one and then link it to the node
		VectorSet(range, 300, 300, 20);
		VectorSubtract(bot->r.currentOrigin, range, mins);
		VectorAdd(bot->r.currentOrigin, range, maxs);

		//get a list of everything in the vicinity of the bot/node
		numTouch = trap_EntitiesInBox(mins, maxs, touch, MAX_GENTITIES);

		//loop through it all
		for (i = 0; i < numTouch; i++)
		{
			obj = &(g_entities[touch[i]]);

			//I think these are the only valid classnames for objectives
			if (bot->client->sess.sessionTeam == TEAM_RED)
			{
				if (!Q_stricmp(obj->classname, "team_CTF_blueflag"))
					break;
			}
			else
			{
				if (!Q_stricmp(obj->classname, "team_CTF_redflag"))
					break;
			}
		}

		if (i == numTouch)	//didn't find an obj
		{
			G_Printf("Error: objective not found at node %d\n", botinfo[bot->s.number].nextNode);
			return qtrue;
		}

		//link the objective we found to the node, so we don't have to search for it again later
		nodes[botinfo[bot->s.number].nextNode].objEntity = obj->s.number;
	}

	//try to find a valid position for the objective
	VectorCopy(obj->r.currentOrigin, origin);
	if ((origin[0] == 0) && (origin[1] == 0) && (origin[2] == 0))
	{
		VectorCopy(obj->s.origin, origin);
		if ((origin[0] == 0) && (origin[1] == 0) && (origin[2] == 0))
		{
			G_Printf("Error: objective at node %d does not appear to have a valid origin!\n", botinfo[bot->s.number].nextNode);
			return qtrue;
		}
	}

	//turn the bot to face it
	VectorSubtract(origin, bot->r.currentOrigin, pos);
	vectoangles(pos, bot->s.angles);

	//and move forward
	if (botinfo[bot->s.number].framesAtNode > 1)
		BotMove(bot, DIR_FORWARD, 0);

	botinfo[bot->s.number].movingAtNode = qtrue;

	return qfalse;
}

/*//////////////////////////////////////////
BotDoFlag
Called when a bot is at a flag node
Searches for a flag in the vicinity of the node, and
directs the bot to touch it if he can take it
*///////////////////////////////////////////
qboolean BotDoFlag(gentity_t *bot)
{
	vec3_t pos, mins, maxs, range, origin;
	int touch[MAX_GENTITIES], numTouch, i;
	gentity_t *flag;

	if ((nodes[botinfo[bot->s.number].nextNode].objEntity != ENTITYNUM_NONE) && (nodes[botinfo[bot->s.number].nextNode].objEntity > MAX_CLIENTS))
		flag = &(g_entities[nodes[botinfo[bot->s.number].nextNode].objEntity]);		//if we found the flag in a previous call
	else
	{
		//set up an area around the bot in which to look for a flag
		VectorSet(range, 300, 300, 20);
		VectorSubtract(bot->r.currentOrigin, range, mins);
		VectorAdd(bot->r.currentOrigin, range, maxs);

		//get all the entities within that area
		numTouch = trap_EntitiesInBox(mins, maxs, touch, MAX_GENTITIES);

		for (i = 0; i < numTouch; i++)	//loop through all the found entities
		{
			flag = &(g_entities[touch[i]]);

			if (!Q_stricmp(flag->classname, "team_WOLF_checkpoint"))	//stop looping once we find the flag
				break;
		}

		if (i == numTouch)		//didn't find a flag
		{
			G_Printf("Error: flag not found at node %d\n", botinfo[bot->s.number].nextNode);
			return qtrue;
		}

		nodes[botinfo[bot->s.number].nextNode].objEntity = flag->s.number;	//so that we don't have to search every frame
	}

	if (nodes[botinfo[bot->s.number].nextNode].objTeam == bot->client->sess.sessionTeam)
	{
		if (flag->count == bot->client->sess.sessionTeam)	//already have the flag
			return qtrue;
	}
	else
	{
		//if it's a flag for the other team, and no one has it yet, skip it
		if (flag->count != OtherTeam(bot->client->sess.sessionTeam))
			return qtrue;
	}

	//oh the joys of different maps behaving differently...
	//first check the r.currentOrigin to see if it's valid
	VectorCopy(flag->r.currentOrigin, origin);
	if ((origin[0] == 0) && (origin[1] == 0) && (origin[2] == 0))
	{
		//if not, check the s.origin
		VectorCopy(flag->s.origin, origin);
		if ((origin[0] == 0) && (origin[1] == 0) && (origin[2] == 0))
		{
			//if they're both invalid...check absmin/max or s.pos.trBase?
			G_Printf("Error: flag at node %d does not appear to have a valid origin!\n", botinfo[bot->s.number].nextNode);
			return qtrue;
		}
	}

	//get the direction to the flag and face it
	VectorSubtract(origin, bot->r.currentOrigin, pos);
	vectoangles(pos, bot->s.angles);

	if (botinfo[bot->s.number].framesAtNode > 1)				//so i'll be facing flag before i move
		BotMove(bot, DIR_FORWARD, 0);							//move towards the flag

	botinfo[bot->s.number].movingAtNode = qtrue;

	return qfalse;												//so this function will get called again
}

