#include "g_local.h"
#include "BotMain.h"

/***************************************************************************************
original code: TeamWhore

filename: BotUtils.c

Description: This file contains utility functions, mainly those not directly related to bots

###################### Function Descriptions ##################################

qboolean MyInfront (gentity_t *self, gentity_t *other)
	The difference between 'MyInfront' and 'infront' is that 'MyInfront' takes into account the 
	bot's skill (skill affects FOV of the bot)

qboolean MyVisible (gentity_t *self, gentity_t *other)
	The difference between MyVisible and 'visible' is that 'MyVisible' takes into account the
	viewheights of the bot and his target, making for more realistic enemy detection

int GetCost(int from, int to)
	Gets the total distance required to move from one node to another node by following
	node paths

int CreatePathAStar(gentity_t *bot, int from, int to, short int *pathlist)
	Uses A* pathfinding to determine the shortest path from one node to another.
	Will likely be modified later

int GetFCost(int to, int num, int parentNum, float *gcost)
	Utility function used by CreatePathAStar to determine the 'F cost' of a given
	node, based on the path used to reach it and its distance from the goal

###############################################################################

-----changes:---------------------
  April 2, 2004
     TeamWhore - created this file and all its contents :P
		moved non-bot functions here from other files
		created CreatePathAStar and GetFCost functions to implement A* pathfinding for bots
		edited Cmd_Push_f to be more accurate
		created new function Cmd_ThrowKnife_f to allow shrub-like knife throwing
		created new function Touch_Knife to handle when a player touches a thrown knife
		implemented Maleficus' PlayerNeedsAmmo function to determine if a player has any use for
			an ammo pack or not
  April/May 2004
	 TeamWhore - created function Cmd_KickGrenade_f to implement grenade kicking of the nearest
			valid grenade with the +salute command.  Possibility of glitches related to this function.
		edited Cmd_ThrowKnife_f and TouchKnife slightly to remove minor bugs and to add new cvar
			functionality
		very small edits to PlayerNeedsAmmo to correct a bug
		added new cvar functionality to Cmd_Push_f to support g_shoveAmount
		created function Cmd_KickAS_f to implement airstrike-canister kicking by LTs
		promoted ClientNumberFromNameMatch to core rather than optional - it's useful!
		added function weapon_smokeGrenade to implement functionality of new smoke grenades
		moved 'optional' mod functions to new file us_cmds.c, since they're not bot-related


****************************************************************************************/

/*////////////////////////////////////////////
MyInfront
my modified version of the built-in 'infront' function
I changed it so that we can have the bots field of vision 'fov'
change based on its skill
returns true if 'other' is in the bot's field of vision
*/////////////////////////////////////////////
qboolean MyInfront (gentity_t *self, gentity_t *other)
{
	vec3_t	vec;
	float	dot;
	vec3_t	forward;

	float	dotConst = 0.0;	//the dotConst values come from this idea - after the dot-product, 90 degrees to either side gives you 0.0
							//directly in front gives you 1.0, and directly behind gives you -1.0 (cosine).  

	switch(botinfo[self->s.number].skill)	//vary the bot's fov by his skill
	{
	case 0:
		dotConst = 0.174;			//160 degrees - like cg_fov 80
		break;
	case 1:
		dotConst = 0.0;				//180 --DEFAULT - like cg_fov 90
		break;
	case 2:
		dotConst = -0.087;			//190 - like cg_fov 95
		break;
	case 3:
		dotConst = -0.216;			//205 - like cg_fov 102.5 - this is closer to a real human's fov, but NOT in a game
		break;
	}
	
	//get the direction in front of the bot
	AngleVectors (self->s.angles, forward, NULL, NULL);
	//get the direction to the 'other'
	VectorSubtract (other->r.currentOrigin, self->r.currentOrigin, vec);
	//make 'vec' (which points from the bot to the 'other') have length 1 (necessary for math...)
	VectorNormalize (vec);
	//essentially get the angle between the bot and the 'other'
	dot = DotProduct (vec, forward);
	//if the 'other' is in the bot's fov, return true (yes, the 'other' is in fron of the bot)
	if (dot > dotConst)
		return qtrue;

	//otherwise return false
	return qfalse;
}


/*////////////////////////////////////////////
MyVisible
This is just a copy of the built-in 'visible' function
but accounting for the players' viewheight.  This lets the
bots see players slightly more often and in a more realistic way
*/////////////////////////////////////////////
qboolean MyVisible (gentity_t *self, gentity_t *other)
{
	vec3_t selfView, otherView;		//basically stores the locations of the players' eyes
	trace_t		tr;					//used in performing the collision info
	gentity_t	*traceEnt;			//holds the entity returned by the trace function

	VectorCopy(self->r.currentOrigin, selfView);		//copy the bot's origin to this variable
	selfView[2] += self->client->ps.viewheight;			//add the bot's viewheight	
	VectorCopy(other->r.currentOrigin, otherView);		//copy the target's origin
	otherView[2] += other->client->ps.viewheight;		//add the target's viewheight

	//check if a shot from the bot's viewheight to the player's viewheight would hit
	trap_Trace (&tr, selfView, NULL, NULL, otherView, self->s.number, MASK_SHOT);

	traceEnt = &g_entities[tr.entityNum];		//set traceEnt to the entity the shot would hit

	if (traceEnt == other)		//if it would hit the player, return true
		return qtrue;

	return qfalse;				//it would hit something other than the player (or nothing) so return false

}

/*///////////////////////////////////////////////////////
GetCost
Attempts to determine the distance a bot would have to move
along a path of nodes to reach one specific node from another
node.  'hops' is the maximum number of links it will follow
before giving up
*////////////////////////////////////////////////////////
int GetCost(int from, int to, int hops)
{
	/*int i, ct=0, link=-1;
	float dist, bestdist=-1;
	float cost = 0;
	vec3_t v;
	trace_t tr;
	node_t *cur;

	if ((from < 0) || (from >= MAX_NODES) || (to < 0) || (to >= MAX_NODES))		//return on bad input
		return -1;

	cur = &(nodes[from]);										//start at the 'from' node

	while (cur->num != to)										//loop until we reach the desired node
	{
		if (ct > hops)
			return -1;					//if we've exceeded the maximum allowed hops, return failure

		if ((cur->num < 0) || (cur->num >= MAX_NODES))	//if this node isn't valid, we have a problem, so return
			return -1;

		for (i = 0; i < cur->numLinks; i++)				//loop through all the links on this node
		{
			if (cur->links[i].target == -1)				//skip any invalid links
				continue;

			if (cur->links[i].flags & PATH_BLOCKED)		//skip any blocked paths
				continue;

			trap_Trace(&tr, nodes[from].pos, NULL, NULL, cur->pos, ENTITYNUM_NONE, MASK_DEADSOLID);
			if (tr.fraction != 1.0)
				continue;

			VectorSubtract(nodes[to].pos, nodes[cur->links[i].target].pos, v);
			dist = VectorLength(v);						//get the distance from the linked node to the endpoint

			if ((dist < bestdist) || (bestdist == -1))	//if it's the best distance we have so far
			{
				bestdist = dist;
				link = i;								//record the node number of the best node yet
			}
		}

		if (link == -1)									//if we didn't find any valid links, we have to return
			return -1;

		cost += cur->links[link].dist;					//add the distance to this linked node to the total
		cur = &(nodes[cur->links[link].target]);		//move to this node and continue the process
		ct++;											//we've moved up a link, so increment our hops counter
	}
	return (int)cost;									//return the total distance between the nodes*/
	return -1;
}

/*////////////////////////////////////////////////////
CreatePathAStar
This function uses the A* pathfinding algorithm to determine the
shortest path between any two nodes.
It's fairly complex, so I'm not really going to explain it much.
Look up A* and binary heaps for more info.
pathlist stores the ideal path between the nodes, in reverse order,
and the return value is the number of nodes in that path
*/////////////////////////////////////////////////////
int CreatePathAStar(gentity_t *bot, int from, int to, short int *pathlist)
{
	//all the data we have to hold...since we can't do dynamic allocation, has to be MAX_NODES
	//we can probably lower this later - eg, the open list should never have more than at most a few dozen items on it
	short int openlist[MAX_NODES+1];	//add 1 because it's a binary heap, and they don't use 0 - 1 is the first used index
	float gcost[MAX_NODES];
	int fcost[MAX_NODES];
	char list[MAX_NODES];	//0 is neither, 1 is open, 2 is closed - char because it's the smallest data type
	short int parent[MAX_NODES];

	short int numOpen = 0;
	short int atNode, temp, newnode=-1;
	qboolean found = qfalse;
	int count = -1;
	float gc;
	int i, u, v, m;
	vec3_t vec;


	//clear out all the arrays
	memset(openlist, 0, sizeof(short int)*(MAX_NODES+1));
	memset(fcost, 0, sizeof(int)*MAX_NODES);
	memset(list, 0, sizeof(char)*MAX_NODES);
	memset(parent, 0, sizeof(short int)*MAX_NODES);
	memset(gcost, -1, sizeof(float)*MAX_NODES);

	//make sure we have valid data before calculating everything
	if ((from == NODE_INVALID) || (to == NODE_INVALID) || (from >= MAX_NODES) || (to >= MAX_NODES) || (from == to))
		return -1;

	openlist[1] = from;		//add the starting node to the open list
	numOpen++;
	gcost[from] = 0;		//its f and g costs are obviously 0
	fcost[from] = 0;

	while (1)
	{
		if (numOpen != 0)	//if there are still items in the open list
		{
			//pop the top item off of the list
			atNode = openlist[1];	
			list[atNode] = 2;		//put the node on the closed list so we don't check it again
			numOpen--;

			openlist[1] = openlist[numOpen+1];	//move the last item in the list to the top position
			v = 1;

			//this while loop reorders the list so that the new lowest fcost is at the top again
			while (1)
			{
				u = v;
				if ((2*u+1) < numOpen)	//if both children exist
				{
					if (fcost[openlist[u]] >= fcost[openlist[2*u]])
						v = 2*u;
					if (fcost[openlist[v]] >= fcost[openlist[2*u+1]])
						v = 2*u+1;
				}
				else
				{
					if ((2*u) < numOpen)	//if only one child exists
					{
						if (fcost[openlist[u]] >= fcost[openlist[2*u]])
							v = 2*u;
					}
				}

				if (u != v)		//if they're out of order, swap this item with its parent
				{
					temp = openlist[u];
					openlist[u] = openlist[v];
					openlist[v] = temp;
				}
				else
					break;
			}

			for (i = 0; i < nodes[atNode].numLinks; i++)	//loop through all the links for this node
			{
				newnode = nodes[atNode].links[i].target;	

				if (nodes[atNode].links[i].flags & PATH_BLOCKED)	//if this path is blocked, skip it
					continue;
				//skip any unreachable nodes
				if ((nodes[newnode].flags & NODE_ALLY_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_BLUE))
					continue;
				if ((nodes[newnode].flags & NODE_AXIS_UNREACHABLE) && (bot->client->sess.sessionTeam == TEAM_RED))
					continue;

				if (list[newnode] == 2)		//if this node is on the closed list, skip it
					continue;

				if (list[newnode] != 1)		//if this node is not already on the open list
				{
					openlist[++numOpen] = newnode;	//add the new node to the open list
					list[newnode] = 1;
					parent[newnode] = atNode;	//record the node's parent

					if (newnode == to)			//if we've found the goal, don't keep computing paths!
						break;			//this will break the 'for' and go all the way to 'if (list[to] == 1)'

					fcost[newnode] = GetFCost(to, newnode, parent[newnode], gcost);	//store it's f cost value

					//this loop re-orders the heap so that the lowest fcost is at the top 
					m = numOpen;
					while (m != 1)	//while this item isn't at the top of the heap already
					{
						if (fcost[openlist[m]] <= fcost[openlist[m/2]])	//if it has a lower fcost than its parent
						{
							temp = openlist[m/2];
							openlist[m/2] = openlist[m];
							openlist[m] = temp;				//swap them
							m /= 2;
						}
						else
							break;
					}
				}
				else		//if this node is already on the open list
				{
					gc = gcost[atNode];
					VectorSubtract(nodes[newnode].pos, nodes[atNode].pos, vec);
					gc += VectorLength(vec);	//calculate what the gcost would be if we reached this node along the current path

					if (gc < gcost[newnode])	//if the new gcost is less (ie, this path is shorter than what we had before)
					{
						parent[newnode] = atNode;	//set the new parent for this node
						gcost[newnode] = gc;		//and the new g cost

						for (i = 1; i < numOpen; i++)	//loop through all the items on the open list
						{
							if (openlist[i] == newnode)	//find this node in the list
							{
								//calculate the new fcost and store it
								fcost[newnode] = GetFCost(to, newnode, parent[newnode], gcost);

								//reorder the list again, with the lowest fcost item on top
								m = i;
								while (m != 1)
								{
									if (fcost[openlist[m]] < fcost[openlist[m/2]])	//if the item has a lower fcost than it's parent
									{
										temp = openlist[m/2];
										openlist[m/2] = openlist[m];
										openlist[m] = temp;				//swap them
										m /= 2;
									}
									else
										break;
								}
								break;	//exit the 'for' loop because we already changed this node
							} //if
						} //for
					} //if (gc < gcost[newnode])
				} //if (list[newnode] != 1) --> else
			} //for (loop through links)
		} //if (numOpen != 0)
		else
		{
			found = qfalse;		//there is no path between these nodes
			break;
		}

		if (list[to] == 1)		//if the destination node is on the open list, we're done
		{
			found = qtrue;
			break;
		}
	} //while (1)

	if (found == qtrue)		//if we found a path
	{
		count = 0;

		temp = to;				//start at the end point
		while (temp != from)	//travel along the path (backwards) until we reach the starting point
		{
			pathlist[count++] = temp;	//add the node to the pathlist and increment the count
			temp = parent[temp];		//move to the parent of this node to continue the path
		}

		pathlist[count++] = from;		//add the beginning node to the end of the pathlist
	}

	return count;	//return the number of nodes in the path, -1 if not found
}

/*///////////////////////////////////////////////////////
GetFCost
Utility function used by A* pathfinding to calculate the
cost to move between nodes towards a goal.  Using the A*
algorithm F = G + H, G here is the distance along the node
paths the bot must travel, and H is the straight-line distance
to the goal node.
Returned as an int because more precision is unnecessary and it
will slightly speed up heap access
*////////////////////////////////////////////////////////
int GetFCost(int to, int num, int parentNum, float *gcost)
{
	float gc = 0;
	float hc = 0;
	vec3_t v;

	if (gcost[num] == -1)
	{
		if (parentNum != -1)
		{
			gc = gcost[parentNum];
			VectorSubtract(nodes[num].pos, nodes[parentNum].pos, v);
			gc += VectorLength(v);
		}
		gcost[num] = gc;
	}
	else
		gc = gcost[num];

	VectorSubtract(nodes[to].pos, nodes[num].pos, v);
	hc = VectorLength(v);

	return (int)(gc + hc);
}
