#include "g_local.h"
#include "us_misc.h"


/***************************************************************************************
original code: TeamWhore

filename: us_svcmds.c

Description: This file contains functions implementing various new server-side commands

###################### Function Descriptions ##################################

void Svcmd_AddNode_f()
	Adds a node to the current nodeset at the server host's current position

void Svcmd_ConnectNodes_f()
	Connects two nodes using the given path flags

void Svcmd_SaveNodes_f()
	Saves the current nodeset to a .nod file

void Svcmd_LoadNodes_f()
	Loads a nodeset from a .nod file and places it into the map

void Svcmd_Entnum_f()
	Displays the entity number and classname of whatever object the server host is
		looking at, as well as the entity number of any func_explosives in the vicinity.

void Svcmd_Kickbot_f()
	Allows the admin to kick multiple bots with one command, by providing a name-match string

void Svcmd_Ignore_f()
	Specifies a player as 'ignored'.  Ignored players may not use text or voice chats, call votes
		or send private messages.  Also implemented as a client vote.

void Svcmd_Unignore_f()
	Removes the 'ignored' specification from a player, so that he may again chat/etc as normal.
		Also implemented as a client vote.

void Svcmd_Shuffle_f()
	Shuffles the teams by randomly placing each active player on a team, while still keeping team
		sizes equal.  Also implemented as a client vote.

###############################################################################

-----changes:---------------------
  April-July, 2004
     TeamWhore - created this file and all its contents
		Created a lot of new functions and moved existing code into them, then moved
			some of those functions to this file, for better organization
		Added a lot of these functions as well, to implement new mod features


****************************************************************************************/


/*/////////////////////////////////////////////////
Svcmd_AddNode_f
Adds a node at the server host's current position in
the map, using the flags he specifies.  The format is
/nd_addnode <flags> [entnum0] [entnum1] [entnum2] [objective flags]
Flags is the type of node, and the entnums are the entity numbers
of any linked objectives.
*//////////////////////////////////////////////////
void Svcmd_AddNode_f()
{
	int flags, objFlags = 0;
	short int entnum[3] = {ENTITYNUM_NONE, ENTITYNUM_NONE, ENTITYNUM_NONE};
	char buf[MAX_TOKEN_CHARS];

	trap_Argv(1, buf, sizeof(buf));
	flags = atoi(buf);

	if (trap_Argc() >= 3)					//these don't have to be specified - they are zeroed if empty
	{
		trap_Argv(2, buf, sizeof(buf));
		entnum[0] = atoi(buf);
	}
	if (trap_Argc() >= 4)
	{
		trap_Argv(3, buf, sizeof(buf));
		entnum[1] = atoi(buf);
	}
	if (trap_Argc() >= 5)
	{
		trap_Argv(4, buf, sizeof(buf));
		entnum[2] = atoi(buf);
	}
	if (trap_Argc() >= 6)
	{
		trap_Argv(5, buf, sizeof(buf));
		objFlags = atoi(buf);
	}

	if (AddNode(g_entities[0].r.currentOrigin, flags, entnum, objFlags))	//add the node
		G_Printf("Added node %d\n", numNodes-1);
	else
		G_Printf("Error: too many nodes\n");
}

/*/////////////////////////////////////////////////
Svcmd_ConnectNodes_f
Connects the two specified nodes with the specified flags.
Format is /nd_connect <from> <to> <flags>
*//////////////////////////////////////////////////
void Svcmd_ConnectNodes_f()
{
	int from, to;
	int flags = 0;
	char buf[MAX_TOKEN_CHARS];

	trap_Argv(1, buf, sizeof(buf));
	from = atoi(buf);
	trap_Argv(2, buf, sizeof(buf));
	to = atoi(buf);
	trap_Argv(3, buf, sizeof(buf));
	flags = atoi(buf);

	if(ConnectNodes(from, to, flags))		//connect the nodes
		G_Printf("Created path from node %d to node %d\n", from, to);
	else
		G_Printf("Error: Invalid nodes, or too many current paths from this node\n");
}

/*//////////////////////////////////////////////////
Svcmd_SaveNodes_f
Saves the current node set to a .nod file with the
specified name.  Do NOT specify the .nod extension,
it is added for you.
*///////////////////////////////////////////////////
void Svcmd_SaveNodes_f()
{
	char buf[MAX_OSPATH], filename[MAX_OSPATH];

	trap_Argv(1, buf, sizeof(buf));

	strcpy(filename, buf);
	strcat(filename, ".nod");				//don't add an extension, it's done for you
	SaveNodes(filename);
}

/*/////////////////////////////////////////////////
Svcmd_LoadNodes_f
Loads a nodeset from the specified .nod file and 
initializes the map with the data.  Do NOT specify the
.nod extension, it is added for you.
*//////////////////////////////////////////////////
void Svcmd_LoadNodes_f()
{
	char buf[MAX_OSPATH], filename[MAX_OSPATH];

	trap_Argv(1, buf, sizeof(buf));

	strcpy(filename, buf);
	strcat(filename, ".nod");
	LoadNodes(filename);
}

/*/////////////////////////////////////////////////
Svcmd_Entnum_f
This is simply a helper function for debugging and creating
nodes.  It displays the entity number and classname of
whatever object you are looking at, as well as the entity
number of any func_explosives that are near what you are 
looking at.  This will almost certainly be removed in
the future.
*//////////////////////////////////////////////////
void Svcmd_Entnum_f()
{
	trace_t tr;
	vec3_t offset, forward, end;
	gentity_t *ent = &g_entities[0];
	int list[10] = {0};
	int i;

	AngleVectors (ent->client->ps.viewangles, forward, NULL, NULL);
	VectorCopy(ent->client->ps.origin, offset);
	offset[2] += ent->client->ps.viewheight;
	VectorMA(offset, 8192, forward, end);

	trap_Trace(&tr, offset, NULL, NULL, end, ent->s.number, (CONTENTS_TRIGGER|CONTENTS_SOLID|CONTENTS_PLAYERCLIP|CONTENTS_BODY|CONTENTS_CORPSE));

	G_Printf("%d %s\n", tr.entityNum, g_entities[tr.entityNum].classname);	//print num and name of whatever we hit

	trap_EntitiesInBox(g_entities[tr.entityNum].r.mins, g_entities[tr.entityNum].r.maxs, list, 10);	//get list of ents near that object

	for (i = 0; i < 10; i++)		//loop through the list looking for a func_explosive
	{
		ent = &g_entities[list[i]];

		if (!ent)
			continue;
		if (!ent->inuse)
			continue;

		if (!Q_stricmp(ent->classname, "func_explosive"))
		{
			G_Printf("%d\n", ent->s.number);				//we found one, so print it and return
			break;
		}
	}
}

/*//////////////////////////////////////////////////
Svcmd_Kickbot_f
A helper function to kick all bots whose name contains
the supplied match string.  For example, if all the bots
have 'bot' in their names, "/kickbot bot" will kick them all.
For kicking a single bot, /kick still works as well.
*///////////////////////////////////////////////////
void Svcmd_Kickbot_f()
{
	int count, i;
	int matchList[MAX_CLIENTS];
	char name[MAX_STRING_CHARS];

	if (trap_Argc() != 2)
	{
		G_Printf("Usage:  kickbot <name>\n");
		return;
	}

	trap_Argv(1, name, sizeof(name));
	count = ClientNumberFromNameMatch(name, matchList);		//this way we can kick multiple bots at once, if we want

	if (count == 0)
	{
		G_Printf("No bots found with that name.\n");
		return;
	}

	//loop through all the matching bots and kick them
	for (i = 0; i < count; i++)
	{
		if (!(g_entities[matchList[i]].r.svFlags & SVF_BOT))	//skip non-bots
			continue;

		ClearBotInfo(matchList[i]);		//clear the botinfo structure
		trap_SendConsoleCommand(EXEC_APPEND, va("clientkick %d\n", matchList[i]));	//do a normal kick command
		numBots--;		//update the number of bots on the server, so that it will be valid on level changes
	}
}


#ifdef UBERSOLDAT_OPTIONAL		//BOTS TW - below are the bot-independent server-side commands
/*///////////////////////////////////////////////////
Svcmd_Ignore_f
A command to allow admins to effectively silence annoying
players, while still allowing them to play.  An ignored
client cannot use any text chats or voice chats, cannot call
votes, and cannot send private messages.  This command is
also implemented through a vote, so that players can vote to
ignore an annoying player even when the admin isn't on the server
*////////////////////////////////////////////////////
void Svcmd_Ignore_f()
{
	int clientNum;
	char buf[5];

	if (trap_Argc() != 2)
	{
		G_Printf("Usage: ignore <clientnum>\n");
		return;
	}

	trap_Argv(1, buf, sizeof(buf));
	clientNum = atoi(buf);

	if ((clientNum < 0) || (clientNum >= MAX_CLIENTS))
	{
		G_Printf("Invalid client number\n");
		return;
	}

	if ((!g_entities[clientNum].client) || (level.clients[clientNum].pers.connected != CON_CONNECTED))
	{
		G_Printf("Client not on server.\n");
		return;
	}

	g_entities[clientNum].ignored = qtrue;		//mark them as ignored
	trap_SendServerCommand(clientNum, "cp \"^3You are now ^1ignored\n\"");	//and tell them, just so they know
}

/*///////////////////////////////////////////////////
Svcmd_Unignore_f
This command simply removes the 'ignore' placed on a
player previously.  It is also implemented as a client
vote.
*////////////////////////////////////////////////////
void Svcmd_Unignore_f()
{
	int clientNum;
	char buf[5];

	if (trap_Argc() != 2)
	{
		G_Printf("Usage: unignore <clientnum>\n");
		return;
	}

	trap_Argv(1, buf, sizeof(buf));
	clientNum = atoi(buf);

	if ((clientNum < 0) || (clientNum >= MAX_CLIENTS))
	{
		G_Printf("Invalid client number.\n");
		return;
	}

	if ((!g_entities[clientNum].client) || (level.clients[clientNum].pers.connected != CON_CONNECTED))
	{
		G_Printf("Client not on server.\n");
		return;
	}

	g_entities[clientNum].ignored = qfalse;
	trap_SendServerCommand(clientNum, "cp \"^3You have been ^2unignored\n\"");	//let them know they can talk again
}

/*///////////////////////////////////////////////////
Svcmd_Shuffle_f
This command will take every active, non-spectating player
and place him randomly on one of the two teams, but still
keeping the number of players on each team the same.  This
effectively 'shuffles' the teams, which can be very useful
if teams are unfairly stacked or balanced.  This is also
implemented as a client vote, so that teams can be shuffled
even when the admin is not present.
*////////////////////////////////////////////////////
void Svcmd_Shuffle_f()
{
	int count=0, tmpCount, i;
	int players[MAX_CLIENTS];

	memset(players, -1, sizeof(players));	//set a default invalid value that we can check against

	for (i = 0; i < MAX_CLIENTS; i++)
	{
		//skip client numbers that aren't used
		if ((!g_entities[i].inuse) || (level.clients[i].pers.connected != CON_CONNECTED))
			continue;

		//ignore spectators
		if ((level.clients[i].sess.sessionTeam != TEAM_RED) && (level.clients[i].sess.sessionTeam != TEAM_BLUE))
			continue;

		players[count] = i;	//record the client numbers of all the active clients
		count++;
	}

	tmpCount = count;	//copy the number of active clients

	//loop through all the active players
	for (i = 0; i < count; i++)
	{
		int j;
			
		do {
			j = (rand() % count);		//pick a random player of those who are active
		} while (players[j] == -1);

		//put every other random choice on allies
		if (i & 1)
			level.clients[players[j]].sess.sessionTeam = TEAM_BLUE;
		else
			level.clients[players[j]].sess.sessionTeam = TEAM_RED;

		//update their info so they actually change teams
		ClientUserinfoChanged(players[j]);
		ClientBegin(players[j]);

		//this is just a way to do quick array resizing - copy end value over used value, and move the endpoint down
		//much faster/easier than shifting each array element
		players[j] = players[tmpCount-1];
		players[tmpCount-1] = -1;
		tmpCount--;
	}

	//let everyone know teams were shuffled
	trap_SendServerCommand(-1, "print \"Teams have been shuffled.\n\"");
	//do a map_restart
	trap_SendConsoleCommand(EXEC_APPEND, va("map_restart 0 %i\n", GS_WARMUP));
}
#endif

